#!/usr/bin/env python

"""
This is a crude script for detecting reference leaks in the C-based cbor2
implementation. It is by no means fool-proof and won't pick up all possible ref
leaks, but it is a reasonable "confidence test" that things aren't horribly
wrong. The script assumes you're in an environment with objgraph and cbor2
installed.

The script outputs a nicely formatted table of the tests run, and the number of
"extra" objects that existed after the tests (indicating a ref-leak), or "-" if
no extra objects existed. The ideal output is obviously "-" in all rows.
"""

import sys
import tracemalloc
from collections import OrderedDict, namedtuple
from datetime import datetime, timedelta, timezone
from decimal import Decimal
from fractions import Fraction

import objgraph

import cbor2

UTC = timezone.utc

TEST_VALUES = [
    # label, kwargs, value
    ("None", {}, None),
    ("10e0", {}, 1),
    ("10e12", {}, 1000000000000),
    ("10e29", {}, 100000000000000000000000000000),
    ("-10e0", {}, -1),
    ("-10e12", {}, -1000000000000),
    ("-10e29", {}, -100000000000000000000000000000),
    ("float1", {}, 1.0),
    ("float2", {}, 3.8),
    ("str", {}, "foo"),
    ("bigstr", {}, "foobarbaz " * 1000),
    ("bytes", {}, b"foo"),
    ("bigbytes", {}, b"foobarbaz\x00" * 1000),
    ("datetime", {"timezone": UTC}, datetime(2019, 5, 9, 22, 4, 5, 123456)),
    ("decimal", {}, Decimal("1.1")),
    ("fraction", {}, Fraction(1, 5)),
    ("intlist", {}, [1, 2, 3]),
    ("bigintlist", {}, [1, 2, 3] * 1000),
    ("strlist", {}, ["foo", "bar", "baz"]),
    ("bigstrlist", {}, ["foo", "bar", "baz"] * 1000),
    ("dict", {}, {"a": 1, "b": 2, "c": 3}),
    ("bigdict", {}, {"a" * i: i for i in range(1000)}),
    ("set", {}, {1, 2, 3}),
    ("bigset", {}, set(range(1000))),
    ("bigdictlist", {}, [{"a" * i: i for i in range(100)}] * 100),
    (
        "objectdict",
        {"timezone": UTC},
        {"name": "Foo", "species": "cat", "dob": datetime(2013, 5, 20), "weight": 4.1},
    ),
    (
        "objectdictlist",
        {"timezone": UTC},
        [{"name": "Foo", "species": "cat", "dob": datetime(2013, 5, 20), "weight": 4.1}] * 100,
    ),
    ("tag", {}, cbor2.CBORTag(1, 1)),
    ("nestedtag", {}, {cbor2.CBORTag(1, 1): 1}),
    ("longstr_128k", {}, "x" * 131072),
    ("longstr_multi_utf8", {}, ("a" * 65535 + "€") * 2),
]

Leaks = namedtuple("Leaks", ("count", "comparison"))
Tests = namedtuple("Tests", ("objgraph", "malloc"))
Result = namedtuple("Result", ("encoding", "decoding", "roundtrip"))


peak = {}


def growth():
    return objgraph.growth(limit=None, peak_stats=peak)


def test_malloc(op):
    count = 0
    start = datetime.now()
    # NOTE: Filter pointing to the op() line in the loop below, because we're
    # only interested in memory allocated by that line. Naturally, if this file
    # is edited, the lineno parameter below must be adjusted!
    only_op = tracemalloc.Filter(True, __file__, lineno=119, all_frames=True)
    tracemalloc.start(10)
    try:
        # Perform a pre-run of op so that any one-time memory allocation
        # (module imports, etc.) don't affect the later diffs
        op()
        before = tracemalloc.take_snapshot().filter_traces([only_op])
        while True:
            count += 1
            op()
            if datetime.now() - start > timedelta(seconds=0.2):
                break
        after = tracemalloc.take_snapshot().filter_traces([only_op])
        diff = after.compare_to(before, "traceback")
        diff = [entry for entry in diff if entry.size_diff > 0]
        return count, diff
    finally:
        tracemalloc.stop()


def test_objgraph(op):
    count = 0
    start = datetime.now()
    # See notes above
    op()
    growth()
    while True:
        count += 1
        op()
        if datetime.now() - start > timedelta(seconds=0.2):
            break
    return count, growth()


def test(op):
    return Tests(Leaks(*test_objgraph(op)), Leaks(*test_malloc(op)))


def format_leaks(result):
    if result.objgraph.comparison:
        num_objs = sum(leak[-1] for leak in result.objgraph.comparison)
        return f"{num_objs} objs (/{result.objgraph.count})"
    elif result.malloc.comparison and (
        result.malloc.count < result.malloc.comparison[0].size_diff
    ):
        # Running the loop always results in *some* memory allocation, but as
        # long as the bytes allocated are less than the number of loops it's
        # unlikely to be an actual leak
        return f"{result.malloc.comparison[0].size_diff} bytes (/{result.malloc.count})"
    else:
        return "-"


def output_table(results):
    # Build table content
    head = ("Test", "Encoding", "Decoding", "Round-trip")
    rows = [head] + [
        (
            label,
            format_leaks(result.encoding),
            format_leaks(result.decoding),
            format_leaks(result.roundtrip),
        )
        for label, result in results.items()
    ]

    # Format table output
    cols = zip(*rows)
    col_widths = [max(len(row) for row in col) for col in cols]
    sep = "".join(
        (
            "+-",
            "-+-".join("-" * width for width in col_widths),
            "-+",
        )
    )
    print(sep)
    print(
        "".join(
            (
                "| ",
                " | ".join(
                    "{value:<{width}}".format(value=value, width=width)
                    for value, width in zip(head, col_widths)
                ),
                " |",
            )
        )
    )
    print(sep)
    for row in rows[1:]:
        print(
            "".join(
                (
                    "| ",
                    " | ".join(
                        "{value:<{width}}".format(value=value, width=width)
                        for value, width in zip(row, col_widths)
                    ),
                    " |",
                )
            )
        )
    print(sep)
    print()
    print(
        """\
There *will* be false positives in the table above. Ignore leaks involving a
tiny number of objects (e.g. 1) or a small number of bytes (e.g. < 8Kb) as such
allocations are quite normal.

In the case of a ref-leak of an object that can reference others (lists, sets,
dicts, or anything with a __dict__), expect to see 100s or 1000s of "objs"
leaked. In the case of a ref-leak of a simple object (int, str, bytes, etc.),
expect to see a few hundred Kb allocated.

If leaks occur across the board, it's likely to be in something universal like
dump/load. If it's restricted to a type, check the encoding and decoding
methods for that type.
"""
    )


def main():
    results = OrderedDict()
    sys.stderr.write("Testing")
    sys.stderr.flush()
    for name, kwargs, value in TEST_VALUES:
        encoded = cbor2.dumps(value, **kwargs)
        results[name] = Result(
            encoding=test(lambda: cbor2.dumps(value, **kwargs)),
            decoding=test(lambda: cbor2.loads(encoded)),
            roundtrip=test(lambda: cbor2.loads(cbor2.dumps(value, **kwargs))),
        )
        sys.stderr.write(".")
        sys.stderr.flush()
    sys.stderr.write("\n")
    sys.stderr.write("\n")
    output_table(results)
    sys.stderr.write("\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
from importlib.metadata import version as get_version

from packaging.version import parse

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
    "sphinx_rtd_theme",
]

templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "cbor2"
author = "Alex Grönholm"
copyright = "2016, " + author

v = parse(get_version(project))
version = v.base_version
release = v.public

language = "en"

exclude_patterns = ["_build"]
pygments_style = "sphinx"
highlight_language = "default"
todo_include_todos = False
autodoc_default_options = {"members": True}
autodoc_mock_imports = ["typing_extensions"]

html_theme = "sphinx_rtd_theme"
htmlhelp_basename = project.replace("-", "") + "doc"

intersphinx_mapping = {"python": ("https://docs.python.org/", None)}



import sys
from collections import OrderedDict, defaultdict

from hypothesis import strategies
from hypothesis.internal.filtering import Ex
from hypothesis.strategies import DrawFn, SearchStrategy

if sys.hexversion < 51314855:
    from cbor2 import frozendict

# Tune these for test run time
MAX_SIZE = 5
MAX_LEAVES = 2

basic_immutable_strategy = strategies.one_of(
    strategies.none(),
    strategies.booleans(),
    strategies.integers(),
    strategies.binary(),
    strategies.text(),
    # nan != nan, so we can't test invariance with it
    strategies.floats(allow_nan=False),
    strategies.decimals(allow_nan=False),
    strategies.datetimes(timezones=strategies.timezones()),
    # strategies.just(undefined),
    strategies.fractions(),
    strategies.uuids(),
    strategies.ip_addresses(),
    # TODO: regex objects for hypothesis
    # MIMEText("foo") != MIMEText("foo"), so we cannot use it to test invariance
    # strategies.text().map(MIMEText),
)
basic_types_strategy = strategies.one_of(
    basic_immutable_strategy,
    strategies.binary().map(bytearray),
)


@strategies.composite
def arbitrary_length_tuple(draw: DrawFn, child_types: SearchStrategy[Ex]) -> tuple[Ex, ...]:
    i = draw(strategies.integers(min_value=0, max_value=MAX_SIZE))
    return tuple(draw(child_types) for _ in range(i))


dict_keys_strategy = strategies.one_of(
    basic_immutable_strategy, arbitrary_length_tuple(basic_immutable_strategy)
)

compound_types_strategy = strategies.recursive(
    strategies.one_of(
        basic_types_strategy,
        strategies.sets(basic_immutable_strategy, max_size=MAX_SIZE),
        strategies.frozensets(basic_immutable_strategy, max_size=MAX_SIZE),
    ),
    lambda children: strategies.one_of(
        strategies.lists(children, max_size=MAX_SIZE),
        # lists and tuples encode to the same, so we can't test invariance with it.
        # Dictionary keys, however, always need to encode to tuples since lists aren't
        # immutable
        strategies.dictionaries(dict_keys_strategy, children, max_size=MAX_SIZE),
        strategies.dictionaries(
            dict_keys_strategy,
            children,
            dict_class=lambda *a: defaultdict(None, *a),  # type: ignore[arg-type]
            max_size=MAX_SIZE,
        ),
        strategies.dictionaries(
            dict_keys_strategy, children, dict_class=OrderedDict, max_size=MAX_SIZE
        ),
        strategies.dictionaries(
            dict_keys_strategy, children, dict_class=frozendict, max_size=MAX_SIZE
        ),
    ),
    max_leaves=MAX_LEAVES,
)



import json
from collections import namedtuple
from operator import attrgetter

from cbor2.decoder import loads as pyloads
from cbor2.encoder import dumps as pydumps

try:
    import _cbor2
except ModuleNotFoundError as e:
    if not str(e).startswith("No module"):
        load_exc = str(e)
    _cbor2 = None

Contender = namedtuple("Contender", "name,dumps,loads")

contenders = []


contenders.append(Contender("json", json.dumps, json.loads))


if _cbor2 is not None:
    contenders.append(Contender("ccbor2", _cbor2.dumps, _cbor2.loads))
else:
    contenders.append(Contender("ppcbor2", pydumps, pyloads))


# See https://github.com/pytest-dev/pytest-cov/issues/418
def pytest_configure(config):
    cov = config.pluginmanager.get_plugin("_cov")
    cov.options.no_cov = True
    if cov.cov_controller:
        cov.cov_controller.pause()


# See https://github.com/ionelmc/pytest-benchmark/issues/48


def pytest_benchmark_group_stats(config, benchmarks, group_by):
    result = {}
    for bench in benchmarks:
        engine, data_kind = bench["param"].split("-")
        group = result.setdefault("{}: {}".format(data_kind, bench["group"]), [])
        group.append(bench)
    return sorted(result.items())


def pytest_generate_tests(metafunc):
    if "contender" in metafunc.fixturenames:
        metafunc.parametrize("contender", contenders, ids=attrgetter("name"))



import binascii
import json
from io import BytesIO, TextIOWrapper
from pathlib import Path

import pytest
from pytest import MonkeyPatch

import cbor2.tool


@pytest.mark.parametrize(
    "value, expected",
    [
        pytest.param((1, 2, 3), [1, 2, 3], id="tuple"),
        pytest.param({b"\x01\x02\x03": "b"}, {"\x01\x02\x03": "b"}, id="byte_key"),
        pytest.param({"dict": {"b": 17}}, {"dict": {"b": 17}}, id="recursion"),
    ],
)
def test_key_to_str(value: object, expected: object) -> None:
    assert cbor2.tool.key_to_str(value) == expected


def test_default() -> None:
    with pytest.raises(TypeError):
        json.dumps(BytesIO(b""), cls=cbor2.tool.DefaultEncoder)


@pytest.mark.parametrize(
    "payload",
    [
        pytest.param("D81CA16162D81CA16161D81D00", id="dict"),
        pytest.param("d81c81d81c830102d81d00", id="list"),
    ],
)
def test_self_referencing(payload: str) -> None:
    decoded = cbor2.loads(binascii.unhexlify(payload))
    with pytest.raises(ValueError, match="Cannot convert self-referential data to JSON"):
        cbor2.tool.key_to_str(decoded)


def test_nonrecursive_ref() -> None:
    payload = "d81c83d81ca26162d81ca16161016163d81d02d81d01d81d01"
    decoded = cbor2.loads(binascii.unhexlify(payload))
    result = cbor2.tool.key_to_str(decoded)
    expected = [
        {"b": {"a": 1}, "c": {"a": 1}},
        {"b": {"a": 1}, "c": {"a": 1}},
        {"b": {"a": 1}, "c": {"a": 1}},
    ]
    assert result == expected


def test_stdin(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    path = tmp_path / "outfile"
    argv = ["-o", str(path)]
    inbuf = TextIOWrapper(BytesIO(binascii.unhexlify("02")))
    with monkeypatch.context() as m:
        m.setattr("sys.argv", [""] + argv)
        m.setattr("sys.stdin", inbuf)
        cbor2.tool.main()
        assert path.read_text() == "2\n"


def test_stdout(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    argv = ["-o", "-"]
    inbuf = TextIOWrapper(BytesIO(binascii.unhexlify("02")))
    outbuf = BytesIO()
    with monkeypatch.context() as m:
        m.setattr("sys.argv", [""] + argv)
        m.setattr("sys.stdin", inbuf)
        m.setattr("sys.stdout", outbuf)
        cbor2.tool.main()


def test_readfrom(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    in_path = tmp_path / "infile"
    out_path = tmp_path / "outfile"
    in_path.write_bytes(binascii.unhexlify("02"))
    argv = ["-o", str(out_path), str(in_path)]
    with monkeypatch.context() as m:
        m.setattr("sys.argv", [""] + argv)
        cbor2.tool.main()
        assert out_path.read_text() == "2\n"


def test_b64(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    out_path = tmp_path / "outfile"
    argv = ["-d", "-o", str(out_path)]
    inbuf = TextIOWrapper(BytesIO(b"oQID"))
    with monkeypatch.context() as m:
        m.setattr("sys.argv", [""] + argv)
        m.setattr("sys.stdin", inbuf)
        cbor2.tool.main()
        assert out_path.read_text() == '{"2": 3}\n'


def test_stream(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    out_path = tmp_path / "outfile"
    argv = ["--sequence", "-o", str(out_path)]
    inbuf = TextIOWrapper(BytesIO(binascii.unhexlify("0203")))
    with monkeypatch.context() as m:
        m.setattr("sys.argv", [""] + argv)
        m.setattr("sys.stdin", inbuf)
        cbor2.tool.main()
        assert out_path.read_text() == "2\n3\n"


def test_embed_bytes(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    out_path = tmp_path / "outfile"
    argv = ["-o", str(out_path)]
    inbuf = TextIOWrapper(BytesIO(binascii.unhexlify("42C2C2")))
    with monkeypatch.context() as m:
        m.setattr("sys.argv", [""] + argv)
        m.setattr("sys.stdin", inbuf)
        cbor2.tool.main()
        assert out_path.read_text() == '"\\\\xc2\\\\xc2"\n'


def test_dtypes_from_file(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    in_path = Path(__file__).with_name("examples.cbor.b64")
    expected = Path(__file__).with_name("examples.json").read_text()
    out_path = tmp_path / "outfile.json"
    argv = ["--sort-keys", "--pretty", "-d", "-o", str(out_path), str(in_path)]
    monkeypatch.setattr("sys.argv", [""] + argv)
    cbor2.tool.main()
    assert out_path.read_text() == expected


def test_ignore_tag(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    out_path = tmp_path / "outfile"
    argv = ["-o", str(out_path), "-i", "6000"]
    inbuf = TextIOWrapper(BytesIO(binascii.unhexlify("D917706548656C6C6F")))
    expected = '"Hello"\n'
    with monkeypatch.context() as m:
        m.setattr("sys.argv", [""] + argv)
        m.setattr("sys.stdin", inbuf)
        cbor2.tool.main()
        assert out_path.read_text() == expected



from random import randint, random
from sys import maxsize

import pytest

composite_object = {
    "words": """
        Lorem ipsum dolor sit amet, consectetur adipiscing
        elit. Mauris adipiscing adipiscing placerat.
        Vestibulum augue augue,
        pellentesque quis sollicitudin id, adipiscing.
        """,
    "list": list(range(200)),
    "dict": {str(i): "a" for i in list(range(200))},
    "int": 100100100,
    "float": 100999.123456,
}

user = {
    "userId": 3381293,
    "age": 213,
    "username": "johndoe",
    "fullname": "John Doe the Second",
    "isAuthorized": True,
    "liked": 31231.31231202,
    "approval": 31.1471,
    "jobs": [1, 2],
    "currJob": None,
}

friends = [user, user, user, user, user, user, user, user]

doubles = []
numbers = []
unicode_strings = []
strings = []
booleans = []
datetimes = []
list_dicts = []
dict_lists = {}

complex_object = [
    [user, friends],
    [user, friends],
    [user, friends],
    [user, friends],
    [user, friends],
    [user, friends],
]

for x in range(256):
    doubles.append(maxsize * random())
    numbers.append(maxsize * random())
    numbers.append(randint(0, maxsize))
    unicode_strings.append(
        "نظام الحكم سلطاني وراثي في الذكو"
        " ر من ذرية السيد تركي بن سعيد بن سلط"
        " ان ويشترط فيمن يختار لولاية الحكم من بينه"
        " م ان يكون مسلما رشيدا عاقلا ًوابنا شرعيا لابوين عمانيين "
    )
    strings.append("A pretty long string which is in a list")
    booleans.append(True)

for y in range(100):
    arrays = []
    list_dicts.append({str(random() * 20): int(random() * 1000000)})

    for x in range(100):
        arrays.append({str(random() * 20): int(random() * 1000000)})
        dict_lists[str(random() * 20)] = arrays


datasets = [
    ("composite object", composite_object),
    ("256 doubles array", doubles),
    ("256 unicode array", unicode_strings),
    ("256 ASCII array", strings),
    ("256 Trues array", booleans),
    ("100 dicts array", list_dicts),
    ("100 arrays dict", dict_lists),
    ("complex object", complex_object),
]


@pytest.mark.benchmark(group="serialize")
@pytest.mark.parametrize("data", [d[1] for d in datasets], ids=[d[0] for d in datasets])
def test_dumps(contender, data, benchmark):
    benchmark(contender.dumps, data)


@pytest.mark.benchmark(group="deserialize")
@pytest.mark.parametrize("data", [d[1] for d in datasets], ids=[d[0] for d in datasets])
def test_loads(contender, data, benchmark):
    data = contender.dumps(data)
    benchmark(contender.loads, data)


@pytest.mark.benchmark(group="roundtrip")
@pytest.mark.parametrize("data", [d[1] for d in datasets], ids=[d[0] for d in datasets])
def test_roundtrip(contender, data, benchmark):
    def roundtrip(d):
        return contender.loads(contender.dumps(d))

    benchmark(roundtrip, data)



import platform
import sys

import pytest

from cbor2 import CBORSimpleValue, CBORTag, undefined

if sys.hexversion < 51314855:
    from cbor2 import frozendict


class TestUndefined:
    def test_bool(self) -> None:
        assert not undefined

    def test_repr(self) -> None:
        assert repr(undefined) == "undefined"

    @pytest.mark.skipif(
        platform.python_implementation() == "PyPy", reason="PyPy does not raise TypeError"
    )
    def test_singleton(self) -> None:
        with pytest.raises(TypeError, match="cannot create 'cbor2.UndefinedType' instances"):
            type(undefined)()


class TestCBORTag:
    def test_bad_args(self) -> None:
        with pytest.raises(TypeError):
            CBORTag("foo", "bar")  # type: ignore[arg-type]

    def test_attr(self) -> None:
        tag = CBORTag(1, "foo")
        assert tag.tag == 1
        assert tag.value == "foo"

    def test_compare(self) -> None:
        tag1 = CBORTag(1, "foo")
        tag2 = CBORTag(1, "foo")
        tag3 = CBORTag(2, "bar")
        tag4 = CBORTag(2, "baz")
        assert tag1 is not tag2
        assert tag1 == tag2
        assert not (tag1 == tag3)
        assert tag1 != tag3
        assert tag3 >= tag2
        assert tag3 > tag2
        assert tag2 < tag3
        assert tag2 <= tag3
        assert tag4 >= tag3
        assert tag4 > tag3
        assert tag3 < tag4
        assert tag3 <= tag4

    def test_compare_unimplemented(self) -> None:
        tag = CBORTag(1, "foo")
        assert not tag == (1, "foo")
        with pytest.raises(TypeError):
            tag <= (1, "foo")

    def test_recursive_repr(self) -> None:
        some_list: list[CBORTag] = []
        tag = CBORTag(1, some_list)
        some_list.append(tag)
        assert repr(tag) == "CBORTag(1, [CBORTag(1, [...])])"

    def test_non_hashable(self) -> None:
        with pytest.raises(RuntimeError, match="This CBORTag is not hashable"):
            hash(CBORTag(1, []))

    def test_repr(self) -> None:
        assert repr(CBORTag(600, "blah")) == "CBORTag(600, 'blah')"


class TestCBORSimpleValue:
    def test_equals(self) -> None:
        tag1 = CBORSimpleValue(1)
        tag2 = CBORSimpleValue(1)
        tag3 = CBORSimpleValue(21)
        tag4 = CBORSimpleValue(99)
        assert tag1 == tag2
        assert tag1 == 1
        assert not tag2 == "21"
        assert tag1 != tag3
        assert tag1 != 21
        assert tag2 != "21"
        assert tag4 > tag1
        assert tag4 >= tag3
        assert 99 <= tag4
        assert 100 > tag4
        assert tag4 <= 100
        assert 2 < tag4
        assert tag4 >= 99
        assert tag1 <= tag4

    def test_ordering(self) -> None:
        randints = [9, 7, 3, 8, 4, 0, 2, 5, 6, 1]
        expected = [CBORSimpleValue(v) for v in range(10)]
        disordered = [CBORSimpleValue(v) for v in randints]
        assert expected == sorted(disordered)
        assert expected == sorted(randints)

    @pytest.mark.parametrize("value", [-1, 24, 31, 256])
    def test_simple_value_out_of_range(self, value: int) -> None:
        with pytest.raises(ValueError, match="simple value out of range"):
            CBORSimpleValue(value)

    def test_repr(self) -> None:
        assert repr(CBORSimpleValue(1)) == "CBORSimpleValue(1)"


class TestFrozendict:
    def test_from_dict(self) -> None:
        d = {1: 2, "foo": "bar"}
        obj = frozendict[int | str, int | str](d)
        assert obj[1] == 2
        assert obj["foo"] == "bar"
        assert obj == d

    def test_from_kwargs(self) -> None:
        obj = frozendict[str, str | int](foo="bar", xyz=123)
        assert obj["foo"] == "bar"
        assert obj["xyz"] == 123

    def test_immutable(self) -> None:
        d = {1: 2, "foo": "bar"}
        obj = frozendict[int | str, int | str](d)
        obj_hash = hash(obj)

        d[1] = 3
        assert obj[1] == 2
        assert hash(obj) == obj_hash

    def test_items(self) -> None:
        obj = frozendict[int, int]({1: 2, 3: 4})
        items = obj.items()
        assert len(items) == 2
        assert list(items) == [(1, 2), (3, 4)]

    def test_keys(self) -> None:
        obj = frozendict[int, int]({1: 2, 3: 4})
        keys = obj.keys()
        assert len(keys) == 2
        assert list(keys) == [1, 3]

    def test_values(self) -> None:
        obj = frozendict[int, int]({1: 2, 3: 4})
        values = obj.values()
        assert len(values) == 2
        assert list(values) == [2, 4]

    def test_get(self) -> None:
        obj = frozendict[int, int]({1: 2})
        assert obj.get(1) == 2
        assert obj.get(3) is None
        assert obj.get(3, "foo") == "foo"

    def test_contains(self) -> None:
        obj = frozendict[int, int]({1: 2})
        assert 1 in obj
        assert 2 not in obj

    def test_len(self) -> None:
        assert len(frozendict({1: 2, 3: 4})) == 2

    def test_repr(self) -> None:
        assert repr(frozendict({1: 2})) == "frozendict({1: 2})"



from __future__ import annotations

import math
import platform
import re
import struct
import sys
from binascii import unhexlify
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from datetime import date, datetime, timedelta, timezone
from decimal import Decimal
from email.message import Message
from fractions import Fraction
from inspect import signature
from io import BytesIO
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
    ip_address,
    ip_interface,
    ip_network,
)
from pathlib import Path
from socket import socketpair
from typing import Any, NoReturn
from uuid import UUID

import pytest
from _pytest.fixtures import FixtureRequest

from cbor2 import (
    CBORDecodeEOF,
    CBORDecodeError,
    CBORDecoder,
    CBORSimpleValue,
    CBORTag,
    dumps,
    load,
    loads,
    shareable_decoder,
    undefined,
)

if sys.hexversion < 51314855:
    from cbor2 import frozendict

DECODER_MAX_DEPTH = signature(loads).parameters["max_depth"].default


@pytest.fixture
def will_overflow() -> bytes:
    """
    Construct an array/string/bytes length which would cause a memory error
    on decode. This should be less than sys.maxsize (the max integer index).
    """
    bit_size = struct.calcsize("P") * 8
    huge_length = 1 << (bit_size - 8)
    return struct.pack("Q", huge_length)


class TestFpAttribute:
    def test_none(self) -> None:
        with pytest.raises(ValueError, match=r"fp must be a readable file-like object"):
            CBORDecoder(None)  # type: ignore[arg-type]

    def test_not_readable(self, tmp_path: Path) -> None:
        # Test for fp not being readable
        with (
            pytest.raises(ValueError, match=r"fp must be a readable file-like object"),
            tmp_path.joinpath("foo.cbor").open("wb") as fp,
        ):
            CBORDecoder(fp)

    def test_delete(self) -> None:
        decoder = CBORDecoder(BytesIO())
        with pytest.raises(AttributeError):
            del decoder.fp


class TestTagHookAttribute:
    def test_callable(self) -> None:
        def tag_hook(tag: CBORTag, immutable: bool) -> object:
            return tag.value

        decoder = CBORDecoder(BytesIO(), tag_hook=tag_hook)
        assert decoder.tag_hook is tag_hook

    def test_not_callable(self) -> None:
        with pytest.raises(TypeError, match="tag_hook must be callable or None"):
            CBORDecoder(BytesIO(), tag_hook="foo")  # type: ignore[arg-type]

    def test_delete(self) -> None:
        decoder = CBORDecoder(BytesIO())
        with pytest.raises(AttributeError):
            del decoder.tag_hook


class TestObjectHookAttribute:
    def test_success(self) -> None:
        def object_hook(value: Mapping[Any, Any], immutable: bool) -> Mapping[Any, Any]:
            return value

        decoder = CBORDecoder(BytesIO(), object_hook=object_hook)
        assert decoder.object_hook is object_hook

    def test_not_callable(self) -> None:
        with pytest.raises(TypeError, match="object_hook must be callable or None"):
            CBORDecoder(BytesIO(), object_hook="foo")  # type: ignore[arg-type]

    def test_delete(self) -> None:
        decoder = CBORDecoder(BytesIO())
        with pytest.raises(AttributeError):
            del decoder.object_hook


class TestStrErrorsAttribute:
    @pytest.mark.parametrize(
        "str_errors", ["strict", "replace", "ignore", "backslashreplace", "surrogateescape"]
    )
    def test_success(self, str_errors: str) -> None:
        decoder = CBORDecoder(BytesIO(), str_errors=str_errors)
        assert decoder.str_errors == str_errors

    def test_invalid(self) -> None:
        with pytest.raises(ValueError, match="invalid str_errors value: 'foo'"):
            CBORDecoder(BytesIO(), str_errors="foo")

    @pytest.mark.parametrize(
        "mode, expected",
        [
            ("strict", None),  # Should raise exception
            ("replace", "hello\ufffdworld"),  # Should replace invalid byte with U+FFFD
        ],
        ids=["strict_mode", "replace_mode"],
    )
    def test_str_errors_handling(self, mode: str, expected: str | None) -> None:
        invalid_utf8 = b"\x6bhello\xffworld"  # \xFF is invalid UTF-8

        if expected is None:
            with pytest.raises(CBORDecodeError, match="error decoding text string"):
                loads(invalid_utf8, str_errors=mode)
        else:
            result = loads(invalid_utf8, str_errors=mode)
            assert result == expected
            assert len(result) == 11
            assert result[5] == "\ufffd"

    def test_str_errors_long_string_over_65536_bytes(self) -> None:
        """Issue #255: str_errors not respected for strings >65536 bytes."""
        # 65537 bytes: 65536 'a' + 1 invalid UTF-8 byte
        payload = unhexlify("7a00010001" + "61" * 65536 + "c3")
        result = loads(payload, str_errors="replace")
        assert len(result) == 65537
        assert result[-1] == "\ufffd"

    def test_str_errors_long_string_invalid_middle(self) -> None:
        """Test str_errors with invalid UTF-8 in the middle of a long string."""
        # 65536 'a' + invalid byte + 65536 'b' = 131073 bytes
        payload = unhexlify("7a00020001" + "61" * 65536 + "c3" + "62" * 65536)
        result = loads(payload, str_errors="replace")
        assert len(result) == 131073
        assert result[65536] == "\ufffd"
        assert result[:65536] == "a" * 65536
        assert result[65537:] == "b" * 65536


def test_allow_indefinite() -> None:
    decoder = CBORDecoder(BytesIO(unhexlify("7f6177ff")), allow_indefinite=False)
    assert not decoder.allow_indefinite
    with pytest.raises(CBORDecodeError, match="encountered indefinite length"):
        decoder.decode()


class TestAllowDuplicateKeys:
    def test_default(self) -> None:
        decoder = CBORDecoder(BytesIO())
        assert decoder.allow_duplicate_keys

    def test_false(self) -> None:
        decoder = CBORDecoder(BytesIO(), allow_duplicate_keys=False)
        assert not decoder.allow_duplicate_keys

    def test_default_allows_duplicates(self) -> None:
        # Definite map {"a": 1, "a": 2} — last value wins
        assert loads(unhexlify("a2616101616102")) == {"a": 2}

    @pytest.mark.parametrize(
        "payload",
        [
            pytest.param("a2616101616102", id="definite"),
            pytest.param("bf616101616102ff", id="indefinite"),
        ],
    )
    @pytest.mark.parametrize("immutable", [False, True])
    def test_raises_on_duplicate(self, payload: str, immutable: bool) -> None:
        with pytest.raises(CBORDecodeError, match="Duplicate map key: 'a'"):
            loads(unhexlify(payload), allow_duplicate_keys=False, immutable=immutable)


def test_readonly_attributes() -> None:
    decoder = CBORDecoder(BytesIO())
    assert decoder.read_size == 4096
    assert decoder.max_depth == DECODER_MAX_DEPTH


def test_read() -> None:
    with BytesIO(b"foobar") as stream:
        decoder = CBORDecoder(stream)
        assert decoder.read(3) == b"foo"
        assert decoder.read(3) == b"bar"

        with pytest.raises(TypeError):
            decoder.read("foo")  # type: ignore[arg-type]

        with pytest.raises(CBORDecodeError):
            decoder.read(10)


def test_non_seekable_fp() -> None:
    sock1, sock2 = socketpair()
    with sock1, sock2:
        receiver = sock2.makefile("rb")
        sock1.sendall(b"\x01\x02extra")
        decoder = CBORDecoder(receiver)
        assert decoder.decode() == 1
        assert decoder.decode() == 2
        assert receiver.read(5) == b"extra"


class TestMaximumDepth:
    def test_default(self) -> None:
        with pytest.raises(
            CBORDecodeError,
            match=f"maximum container nesting depth \\({DECODER_MAX_DEPTH}\\) exceeded",
        ):
            loads(b"\x81" * (DECODER_MAX_DEPTH + 1) + b"\x80")

    def test_explicit(self) -> None:
        with pytest.raises(
            CBORDecodeError, match=r"maximum container nesting depth \(9\) exceeded"
        ):
            loads(b"\x81" * 10 + b"\x80", max_depth=9)


def test_stream_position_after_decode() -> None:
    """Test that the stream position is exactly at the end of the decoded CBOR value."""
    # CBOR: integer 1, followed by non-CBOR data ("extra")
    stream = BytesIO(b"\x01extra")
    assert load(stream) == 1
    # Stream position should be exactly at the end of CBOR data
    assert stream.tell() == 1
    # Should be able to read the extra data
    assert stream.read() == b"extra"


@pytest.mark.parametrize(
    "payload, expected",
    [
        ("00", 0),
        ("01", 1),
        ("0a", 10),
        ("17", 23),
        ("1818", 24),
        ("1819", 25),
        ("1864", 100),
        ("1903e8", 1000),
        ("1a000f4240", 1000000),
        ("1b000000e8d4a51000", 1000000000000),
        ("1bffffffffffffffff", 18446744073709551615),
        ("c249010000000000000000", 18446744073709551616),
        ("3bffffffffffffffff", -18446744073709551616),
        ("c349010000000000000000", -18446744073709551617),
        ("20", -1),
        ("29", -10),
        ("3863", -100),
        ("3903e7", -1000),
    ],
)
def test_integer(payload: str, expected: int) -> None:
    decoded = loads(unhexlify(payload))
    assert decoded == expected


def test_invalid_integer_subtype() -> None:
    with pytest.raises(CBORDecodeError) as exc:
        loads(b"\x1c")
        assert str(exc.value).endswith("unknown unsigned integer subtype 0x1c")
        assert isinstance(exc, ValueError)


@pytest.mark.parametrize(
    "payload, expected",
    [
        ("f90000", 0.0),
        ("f98000", -0.0),
        ("f93c00", 1.0),
        ("fb3ff199999999999a", 1.1),
        ("f93e00", 1.5),
        ("f97bff", 65504.0),
        ("fa47c35000", 100000.0),
        ("fa7f7fffff", 3.4028234663852886e38),
        ("fb7e37e43c8800759c", 1.0e300),
        ("f90001", 5.960464477539063e-8),
        ("f90400", 0.00006103515625),
        ("f9c400", -4.0),
        ("fbc010666666666666", -4.1),
        ("f97c00", float("inf")),
        ("f9fc00", float("-inf")),
        ("fa7f800000", float("inf")),
        ("faff800000", float("-inf")),
        ("fb7ff0000000000000", float("inf")),
        ("fbfff0000000000000", float("-inf")),
    ],
)
def test_float(payload: str, expected: float) -> None:
    decoded = loads(unhexlify(payload))
    assert decoded == expected


@pytest.mark.parametrize("payload", ["f97e00", "fa7fc00000", "fb7ff8000000000000"])
def test_float_nan(payload: str) -> None:
    decoded = loads(unhexlify(payload))
    assert math.isnan(decoded)


@pytest.fixture(
    params=[
        pytest.param(("f4", False), id="false"),
        pytest.param(("f5", True), id="true"),
        pytest.param(("f6", None), id="null"),
        pytest.param(("f7", "undefined"), id="undefined"),
    ],
)
def special_values(request: FixtureRequest) -> tuple[str, Any]:
    payload, expected = request.param
    if expected == "undefined":
        expected = undefined

    return payload, expected


def test_special(special_values: tuple[str, Any]) -> None:
    payload, expected = special_values
    decoded = loads(unhexlify(payload))
    assert decoded is expected


@pytest.mark.parametrize(
    "payload, expected",
    [
        pytest.param("40", b"", id="blank"),
        pytest.param("4401020304", b"\x01\x02\x03\x04", id="short"),
        pytest.param("5a00011170" + "12" * 70000, b"\x12" * 70000, id="long"),
    ],
)
def test_binary(payload: str, expected: bytes) -> None:
    decoded = loads(unhexlify(payload))
    assert decoded == expected


@pytest.mark.parametrize(
    "payload, expected",
    [
        ("60", ""),
        ("6161", "a"),
        ("6449455446", "IETF"),
        ("62225c", '"\\'),
        ("62c3bc", "\u00fc"),
        ("63e6b0b4", "\u6c34"),
        pytest.param("7a00010001" + "61" * 65535 + "c3b6", "a" * 65535 + "ö", id="split_unicode"),
    ],
)
def test_string(payload: str, expected: str) -> None:
    decoded = loads(unhexlify(payload))
    assert decoded == expected


@pytest.mark.parametrize(
    "payload",
    [
        pytest.param("6198", id="short"),
        pytest.param("7a00010000" + "61" * 65535 + "c3", id="long"),
        pytest.param("7f6198ff", id="indefinite"),
    ],
)
def test_string_invalid_utf8(payload: str) -> None:
    with pytest.raises(CBORDecodeError, match="error decoding text string") as exc:
        loads(unhexlify(payload))

    assert isinstance(exc.value.__cause__, UnicodeDecodeError)


def test_string_oversized() -> None:
    with pytest.raises(CBORDecodeEOF, match="premature end of stream"):
        loads(unhexlify("aeaeaeaeaeaeaeaeae0108c29843d90100d8249f0000aeaeffc26ca799"))


def test_string_issue_264_multiple_chunks_utf8_boundary() -> None:
    """Test for Issue #264: UTF-8 characters split across multiple 65536-byte chunk boundaries."""
    import struct

    # Construct: 65535 'a' + '€' (3 bytes) + 65533 'b' + '€' (3 bytes) + 100 'd'
    # Total: 131174 bytes, which spans 3 chunks (65536 + 65536 + 102)
    total_bytes = 65535 + 3 + 65533 + 3 + 100

    payload = b"\x7a" + struct.pack(">I", total_bytes)  # major type 3, 4-byte length
    payload += b"a" * 65535
    payload += "€".encode()  # U+20AC: E2 82 AC
    payload += b"b" * 65533
    payload += "€".encode()
    payload += b"d" * 100

    expected = "a" * 65535 + "€" + "b" * 65533 + "€" + "d" * 100

    result = loads(payload)
    assert result == expected
    assert len(result) == 131170  # 65535 + 1 + 65533 + 1 + 100 characters


@pytest.mark.parametrize(
    "payload, expected",
    [
        ("80", []),
        ("83010203", [1, 2, 3]),
        ("8301820203820405", [1, [2, 3], [4, 5]]),
        (
            "98190102030405060708090a0b0c0d0e0f101112131415161718181819",
            list(range(1, 26)),
        ),
    ],
)
def test_array(payload: str, expected: list[Any]) -> None:
    decoded = loads(unhexlify(payload))
    assert decoded == expected


@pytest.mark.parametrize("payload, expected", [("a0", {}), ("a201020304", {1: 2, 3: 4})])
def test_map(payload: str, expected: dict[int, Any]) -> None:
    decoded = loads(unhexlify(payload))
    assert decoded == expected


def test_nested_map() -> None:
    # Regression test for #295
    result = loads(unhexlify("a101a102a103f6"))
    assert result == {1: {2: {3: None}}}
    assert str(result) == "{1: {2: {3: None}}}"


@pytest.mark.parametrize(
    "payload, expected",
    [
        ("a26161016162820203", {"a": 1, "b": [2, 3]}),
        ("826161a161626163", ["a", {"b": "c"}]),
        (
            "a56161614161626142616361436164614461656145",
            {"a": "A", "b": "B", "c": "C", "d": "D", "e": "E"},
        ),
    ],
)
def test_mixed_array_map(payload: str, expected: dict[str, Any]) -> None:
    decoded = loads(unhexlify(payload))
    assert decoded == expected


@pytest.mark.parametrize(
    "payload, expected",
    [
        ("5f42010243030405ff", b"\x01\x02\x03\x04\x05"),
        ("7f657374726561646d696e67ff", "streaming"),
        ("9fff", []),
        ("9f018202039f0405ffff", [1, [2, 3], [4, 5]]),
        ("9f01820203820405ff", [1, [2, 3], [4, 5]]),
        ("83018202039f0405ff", [1, [2, 3], [4, 5]]),
        ("83019f0203ff820405", [1, [2, 3], [4, 5]]),
        (
            "9f0102030405060708090a0b0c0d0e0f101112131415161718181819ff",
            list(range(1, 26)),
        ),
        ("bf61610161629f0203ffff", {"a": 1, "b": [2, 3]}),
        ("826161bf61626163ff", ["a", {"b": "c"}]),
        ("bf6346756ef563416d7421ff", {"Fun": True, "Amt": -2}),
        ("d901029f010203ff", {1, 2, 3}),
    ],
)
def test_streaming(payload: str, expected: object) -> None:
    decoded = loads(unhexlify(payload))
    assert decoded == expected


@pytest.mark.parametrize(
    "payload",
    [
        "5f42010200",
        "7f63737472a0",
    ],
)
def test_bad_streaming_strings(payload: str) -> None:
    with pytest.raises(
        CBORDecodeError,
        match=r"non-(byte|text) string \(major type \d\) found in indefinite length (byte|text) string",
    ):
        loads(unhexlify(payload))


@pytest.mark.parametrize(
    "payload, value",
    [
        ("e0", 0),
        ("e2", 2),
        ("f3", 19),
        ("f820", 32),
    ],
)
def test_simple_value(payload: str, value: int) -> None:
    wrapped = CBORSimpleValue(value)
    decoded = loads(unhexlify(payload))
    assert decoded == value
    assert decoded == wrapped


def test_simple_val_as_key() -> None:
    decoded = loads(unhexlify("A1F86301"))
    assert decoded == {CBORSimpleValue(99): 1}


@pytest.mark.parametrize("payload", ["f800", "f817", "f81f"])
def test_invalid_simple_value(payload: str) -> None:
    with pytest.raises(CBORDecodeError, match="invalid two-byte sequence for simple value"):
        loads(unhexlify(payload))


#
# Tests for extension tags
#


@pytest.mark.parametrize(
    "payload, expected",
    [
        pytest.param(
            "d903ec6a323031332d30332d3231",
            date(2013, 3, 21),
            id="date/string",
        ),
        pytest.param(
            "d8641945e8",
            date(2018, 12, 31),
            id="date/timestamp",
        ),
    ],
)
def test_date(payload: str, expected: date) -> None:
    decoded = loads(unhexlify(payload))
    assert decoded == expected


@pytest.mark.parametrize(
    "payload, expected",
    [
        pytest.param(
            "c074323031332d30332d32315432303a30343a30305a",
            datetime(2013, 3, 21, 20, 4, 0, tzinfo=timezone.utc),
            id="datetime/utc",
        ),
        pytest.param(
            "c0781b323031332d30332d32315432303a30343a30302e3338303834315a",
            datetime(2013, 3, 21, 20, 4, 0, 380841, tzinfo=timezone.utc),
            id="datetime+micro/utc",
        ),
        pytest.param(
            "c07819323031332d30332d32315432323a30343a30302b30323a3030",
            datetime(2013, 3, 21, 22, 4, 0, tzinfo=timezone(timedelta(hours=2))),
            id="datetime/eet",
        ),
        pytest.param(
            "c11a514b67b0",
            datetime(2013, 3, 21, 20, 4, 0, tzinfo=timezone.utc),
            id="timestamp/utc",
        ),
        pytest.param(
            "c11a514b67b0",
            datetime(2013, 3, 21, 22, 4, 0, tzinfo=timezone(timedelta(hours=2))),
            id="timestamp/eet",
        ),
    ],
)
def test_datetime(payload: str, expected: datetime) -> None:
    decoded = loads(unhexlify(payload))
    assert decoded == expected


@pytest.mark.parametrize(
    "payload, expected",
    [
        (
            b"\xc0\x78\x162018-08-02T07:00:59.1Z",
            datetime(2018, 8, 2, 7, 0, 59, 100000, tzinfo=timezone.utc),
        ),
        (
            b"\xc0\x78\x172018-08-02T07:00:59.01Z",
            datetime(2018, 8, 2, 7, 0, 59, 10000, tzinfo=timezone.utc),
        ),
        (
            b"\xc0\x78\x182018-08-02T07:00:59.001Z",
            datetime(2018, 8, 2, 7, 0, 59, 1000, tzinfo=timezone.utc),
        ),
        (
            b"\xc0\x78\x192018-08-02T07:00:59.0001Z",
            datetime(2018, 8, 2, 7, 0, 59, 100, tzinfo=timezone.utc),
        ),
        (
            b"\xc0\x78\x1a2018-08-02T07:00:59.00001Z",
            datetime(2018, 8, 2, 7, 0, 59, 10, tzinfo=timezone.utc),
        ),
        (
            b"\xc0\x78\x1b2018-08-02T07:00:59.000001Z",
            datetime(2018, 8, 2, 7, 0, 59, 1, tzinfo=timezone.utc),
        ),
        (
            b"\xc0\x78\x1c2018-08-02T07:00:59.0000001Z",
            datetime(2018, 8, 2, 7, 0, 59, 0, tzinfo=timezone.utc),
        ),
    ],
)
def test_datetime_secfrac(payload: bytes, expected: datetime) -> None:
    assert loads(payload) == expected


def test_datetime_secfrac_overflow() -> None:
    decoded = loads(b"\xc0\x78\x2c2018-08-02T07:00:59.100500999999999999+00:00")
    assert decoded == datetime(2018, 8, 2, 7, 0, 59, 100500, tzinfo=timezone.utc)
    decoded = loads(b"\xc0\x78\x2c2018-08-02T07:00:59.999999999999999999+00:00")
    assert decoded == datetime(2018, 8, 2, 7, 0, 59, 999999, tzinfo=timezone.utc)


def test_datetime_invalid_string() -> None:
    with pytest.raises(CBORDecodeError) as excinfo:
        loads(unhexlify("c06b303030302d3132332d3031"))

    assert isinstance(excinfo.value.__cause__, ValueError)
    assert str(excinfo.value.__cause__) == "Invalid isoformat string: '0000-123-01'"


def test_datetime_overflow() -> None:
    with pytest.raises(CBORDecodeError) as excinfo:
        loads(unhexlify("c11b9b9b9b0000000000"))

    assert isinstance(excinfo.value.__cause__, OverflowError)


def test_datetime_value_too_large() -> None:
    with pytest.raises(CBORDecodeError) as excinfo:
        loads(unhexlify("c11b1616161616161616161616161616"))

    assert excinfo.value.__cause__ is not None


def test_datetime_date_out_of_range() -> None:
    with pytest.raises(CBORDecodeError) as excinfo:
        loads(unhexlify("a6c11b00002401001b000000000000ff00"))

    if platform.system() == "Windows":
        cause_exc_class: type[Exception] = OSError
    elif sys.maxsize == 2147483647:
        cause_exc_class = OverflowError
    else:
        cause_exc_class = ValueError

    assert isinstance(excinfo.value.__cause__, cause_exc_class)


def test_datetime_timezone() -> None:
    decoded = loads(b"\xc0\x78\x192018-08-02T07:00:59+00:30")
    assert decoded == datetime(2018, 8, 2, 7, 0, 59, tzinfo=timezone(timedelta(minutes=30)))
    decoded = loads(b"\xc0\x78\x192018-08-02T07:00:59-00:30")
    assert decoded == datetime(2018, 8, 2, 7, 0, 59, tzinfo=timezone(timedelta(minutes=-30)))
    decoded = loads(b"\xc0\x78\x192018-08-02T07:00:59+01:30")
    assert decoded == datetime(2018, 8, 2, 7, 0, 59, tzinfo=timezone(timedelta(minutes=90)))
    decoded = loads(b"\xc0\x78\x192018-08-02T07:00:59-01:30")
    assert decoded == datetime(2018, 8, 2, 7, 0, 59, tzinfo=timezone(timedelta(minutes=-90)))


def test_positive_bignum() -> None:
    # Example from RFC 8949 section 3.4.3.
    decoded = loads(unhexlify("c249010000000000000000"))
    assert decoded == 18446744073709551616


def test_negative_bignum() -> None:
    decoded = loads(unhexlify("c349010000000000000000"))
    assert decoded == -18446744073709551617


def test_fraction() -> None:
    decoded = loads(unhexlify("c48221196ab3"))
    assert decoded == Decimal("273.15")


def test_decimal_precision() -> None:
    decoded = loads(unhexlify("c482384dc252011f1fe37d0c70ff50456ba8b891997b07d6"))
    assert decoded == Decimal("9.7703426561852468194804075821069770622934E-38")


def test_bigfloat() -> None:
    decoded = loads(unhexlify("c5822003"))
    assert decoded == Decimal("1.5")


@pytest.mark.parametrize(
    "payload, expected",
    [
        ("d9a7f882f90000f90000", 0.0j),
        ("d9a7f882fb0000000000000000fb0000000000000000", 0.0j),
        ("d9a7f882f98000f98000", -0.0j),
        ("d9a7f882f90000f93c00", 1.0j),
        ("d9a7f882fb0000000000000000fb3ff199999999999a", 1.1j),
        ("d9a7f882f93e00f93e00", 1.5 + 1.5j),
        ("d9a7f882f97bfff97bff", 65504.0 + 65504.0j),
        ("d9a7f882fa47c35000fa47c35000", 100000.0 + 100000.0j),
        ("d9a7f882f90000fb7e37e43c8800759c", 1.0e300j),
        ("d9a7f882f90000f90001", 5.960464477539063e-8j),
        ("d9a7f882f90000f90400", 0.00006103515625j),
        ("d9a7f882f90000f9c400", -4.0j),
        ("d9a7f882f90000fbc010666666666666", -4.1j),
        ("d9a7f882f90000f97c00", complex(0.0, float("inf"))),
        ("d9a7f882f97c00f90000", complex(float("inf"), 0.0)),
        ("d9a7f882f90000f9fc00", complex(0.0, float("-inf"))),
        ("d9a7f882f90000fa7f800000", complex(0.0, float("inf"))),
        ("d9a7f882f90000faff800000", complex(0.0, float("-inf"))),
        ("d9a7f882f97e00fb0000000000000000", complex(float("nan"), 0.0)),
        ("d9a7f882fb0000000000000000f97e00", complex(0.0, float("nan"))),
        ("d9a7f882f97e00f97e00", complex(float("nan"), float("nan"))),
    ],
)
def test_complex(payload: str, expected: complex) -> None:
    decoded = loads(unhexlify(payload))
    if math.isnan(expected.real):
        assert math.isnan(decoded.real)
    else:
        assert expected.real == decoded.real

    if math.isnan(expected.imag):
        assert math.isnan(decoded.imag)
    else:
        assert expected.imag == decoded.imag


def test_rational() -> None:
    decoded = loads(unhexlify("d81e820205"))
    assert decoded == Fraction(2, 5)


def test_rational_invalid_iterable() -> None:
    with pytest.raises(CBORDecodeError, match="error decoding rational") as exc_info:
        loads(unhexlify("d81e01"))

    assert isinstance(exc_info.value.__cause__, TypeError)
    assert str(exc_info.value.__cause__) == "input value must be an array"


def test_rational_zero_denominator() -> None:
    with pytest.raises(CBORDecodeError, match="error decoding rational") as exc:
        loads(unhexlify("d81e820100"))

    assert isinstance(exc.value.__cause__, ZeroDivisionError)


def test_regex() -> None:
    decoded = loads(unhexlify("d8236d68656c6c6f2028776f726c6429"))
    expr = re.compile("hello (world)")
    assert decoded == expr


def test_regex_unbalanced_parentheses() -> None:
    with pytest.raises(CBORDecodeError, match="error decoding regular expression") as exc:
        loads(unhexlify("d8236c68656c6c6f2028776f726c64"))

    assert isinstance(exc.value.__cause__, re.error)


def test_mime() -> None:
    decoded = loads(
        unhexlify(
            "d824787b436f6e74656e742d547970653a20746578742f706c61696e3b20636861727365743d2269736f"
            "2d383835392d3135220a4d494d452d56657273696f6e3a20312e300a436f6e74656e742d5472616e7366"
            "65722d456e636f64696e673a2071756f7465642d7072696e7461626c650a0a48656c6c6f203d41347572"
            "6f"
        )
    )
    assert isinstance(decoded, Message)
    assert decoded.get_payload() == "Hello =A4uro"


def test_mime_invalid_type() -> None:
    with pytest.raises(CBORDecodeError, match="error decoding MIME message") as exc:
        loads(unhexlify("d82401"))

    assert isinstance(exc.value.__cause__, TypeError)


def test_uuid() -> None:
    decoded = loads(unhexlify("d825505eaffac8b51e480581277fdcc7842faf"))
    assert decoded == UUID(hex="5eaffac8b51e480581277fdcc7842faf")


def test_uuid_invalid_length() -> None:
    with pytest.raises(CBORDecodeError, match="error decoding UUID") as exc:
        loads(unhexlify("d8254f5eaffac8b51e480581277fdcc7842f"))

    assert isinstance(exc.value.__cause__, ValueError)


def test_uuid_invalid_type() -> None:
    with pytest.raises(CBORDecodeError, match="error decoding UUID") as exc:
        loads(unhexlify("d82501"))

    assert isinstance(exc.value.__cause__, TypeError)


@pytest.mark.parametrize(
    "payload, expected",
    [
        pytest.param("d83444c0000201", IPv4Address("192.0.2.1"), id="ipv4addr"),
        pytest.param("d83482181843c00002", IPv4Network("192.0.2.0/24"), id="ipv4net"),
        pytest.param("d8348244c00002011818", IPv4Interface("192.0.2.1/24"), id="ipv4if"),
        pytest.param(
            "d8365020010db81234deedbeefcafefacefeed",
            IPv6Address("2001:0db8:1234:deed:beef:cafe:face:feed"),
            id="ipv6addr",
        ),
        pytest.param(
            "d8368218304620010db81234",
            IPv6Network("2001:db8:1234::/48"),
            id="ipv6net",
        ),
        pytest.param(
            "d8368350fe8000000000020202fffffffe03030318404465746830",
            IPv6Interface("fe80::202:2ff:ffff:fe03:303%eth0/64"),
            id="ipv6if_str_zoneid",
        ),
        pytest.param(
            "d8368350fe8000000000020202fffffffe030303184002",
            IPv6Interface("fe80::202:2ff:ffff:fe03:303%2/64"),
            id="ipv6if_num_zoneid",
        ),
    ],
)
def test_ipaddress(payload: bytes, expected: Any) -> None:
    assert loads(unhexlify(payload)) == expected


class TestDeprecatedIPAddress:
    @pytest.mark.parametrize(
        "payload, expected",
        [
            pytest.param("d9010444c00a0a01", ip_address("192.10.10.1"), id="ipv4"),
            pytest.param(
                "d901045020010db885a3000000008a2e03707334",
                ip_address("2001:db8:85a3::8a2e:370:7334"),
                id="ipv6",
            ),
            pytest.param(
                "d9010446010203040506", CBORTag(260, b"\x01\x02\x03\x04\x05\x06"), id="mac"
            ),
        ],
    )
    def test_valid(self, payload: str, expected: Any) -> None:
        assert loads(unhexlify(payload)) == expected

    @pytest.mark.parametrize("payload", ["d9010443c00a0a", "d9010401"])
    def test_invalid(self, payload: str) -> None:
        with pytest.raises(CBORDecodeError, match="error decoding IP address"):
            loads(unhexlify(payload))


class TestDeprecatedIPNetwork:
    @pytest.mark.parametrize(
        "payload, expected",
        [
            pytest.param("D90105A144C0A800001818", ip_network("192.168.0.0/24"), id="ipv4_net"),
            pytest.param("d90105a144c0a800641818", ip_interface("192.168.0.100/24"), id="ipv4_if"),
            pytest.param(
                "d90105a15020010db885a3000000008a2e000000001860",
                ip_network("2001:db8:85a3:0:0:8a2e::/96"),
                id="ipv6_net",
            ),
        ],
    )
    def test_valid(self, payload: str, expected: Any) -> None:
        assert loads(unhexlify(payload)) == expected

    @pytest.mark.parametrize(
        "payload, pattern",
        [
            pytest.param(
                "d90105a244c0a80064181844c0a800001818",
                "invalid input map length for IP network: 2",
                id="length",
            ),
            pytest.param(
                "d90105a144c0a80064420102",
                r"invalid mask length for IP network: b'\\x01\\x02'",
                id="mask",
            ),
        ],
    )
    def test_invalid(self, payload: str, pattern: str) -> None:
        with pytest.raises(CBORDecodeError, match=pattern):
            loads(unhexlify(payload))


class TestSharedReference:
    def test_bad_reference(self) -> None:
        with pytest.raises(CBORDecodeError, match="shared reference 5 not found"):
            loads(unhexlify("d81d05"))

    def test_uninitialized(self) -> None:
        with pytest.raises(CBORDecodeError, match="shared value 0 has not been initialized"):
            loads(unhexlify("D90102D81C8181D81D00"))

    def test_immutable(self) -> None:
        # a = (1, 2, 3)
        # b = ((a, a), a)
        # data = dumps(set(b))
        decoded = loads(unhexlify("d90102d81c82d81c82d81c83010203d81d02d81d02"))
        a = [item for item in decoded if len(item) == 3][0]
        b = [item for item in decoded if len(item) == 2][0]
        assert decoded == {(a, a), a}
        assert b[0] is a
        assert b[1] is a

    def test_cyclic_array(self) -> None:
        decoded = loads(unhexlify("d81c81d81d00"))
        assert decoded == [decoded]

    def test_cyclic_map(self) -> None:
        decoded = loads(unhexlify("d81ca100d81d00"))
        assert decoded == {0: decoded}

    def test_nested_shareable_in_array(self) -> None:
        decoded = loads(unhexlify("82d81c82d81c61616162d81d00"))
        assert decoded == [["a", "b"], ["a", "b"]]
        assert decoded[0] is decoded[1]


class TestStringReference:
    def test_string_ref(self) -> None:
        decoded = loads(unhexlify("d9010085656669727374d81900667365636f6e64d81900d81901"))
        assert isinstance(decoded, list)
        assert decoded == ["first", "first", "second", "first", "second"]

    def test_ref_outside_of_namespace(self) -> None:
        with pytest.raises(CBORDecodeError, match="string reference outside of namespace$"):
            loads(unhexlify("85656669727374d81900667365636f6e64d81900d81901"))

    def test_invalid_string_ref(self) -> None:
        with pytest.raises(CBORDecodeError, match="string reference 3 not found$"):
            loads(unhexlify("d9010086656669727374d81900667365636f6e64d81900d81901d81903"))

    def test_string_ref_long_text_string(self) -> None:
        """Regression test for #308."""
        big_string = "a" * 65537
        payload = (
            b"\xd9\x01\x00"  # tag 256 (string namespace)
            b"\x82"  # array of 2
            b"\x7a"
            + struct.pack(">I", 65537)  # text string, 4-byte length
            + big_string.encode("ascii")  # 65537 bytes of 'a'
            + b"\xd8\x19\x00"  # tag 25, value 0 (string reference to index 0)
        )
        assert loads(payload) == [big_string, big_string]


@pytest.mark.parametrize(
    "payload, expected",
    [
        pytest.param("d9d9f71903e8", 1000, id="self_describe_cbor+int"),
        pytest.param(
            "d9d9f7c249010000000000000000",
            18446744073709551616,
            id="self_describe_cbor+positive_bignum",
        ),
    ],
)
def test_self_describe_cbor(payload: str, expected: object) -> None:
    assert loads(unhexlify(payload)) == expected


def test_unhandled_tag() -> None:
    """
    Test that a tag is simply ignored and its associated value returned if there is no special
    handling available for it.

    """
    decoded = loads(unhexlify("d917706548656c6c6f"))
    assert decoded == CBORTag(6000, "Hello")


def test_premature_end_of_stream() -> None:
    """
    Test that the decoder detects a situation where read() returned fewer than expected bytes.

    """
    with pytest.raises(CBORDecodeError) as exc:
        loads(unhexlify("437879"))
        exc.match(r"premature end of stream \(expected to read 3 bytes, got 2 instead\)")
        assert isinstance(exc, EOFError)


def test_tag_hook() -> None:
    def reverse(tag: CBORTag, immutable: bool) -> Any:
        assert tag.tag == 6000
        return tag.value[::-1]

    decoded = loads(unhexlify("d917706548656c6c6f"), tag_hook=reverse)
    assert decoded == "olleH"


def test_object_hook() -> None:
    class DummyType:
        def __init__(self, state: Mapping[Any, Any], immutable: bool) -> None:
            self.state = state

    payload = unhexlify("A2616103616205")
    decoded = loads(payload, object_hook=DummyType)
    assert isinstance(decoded, DummyType)
    assert decoded.state == {"a": 3, "b": 5}


def test_object_hook_exception() -> None:
    def object_hook(data: Mapping[Any, Any], immutable: bool) -> NoReturn:
        raise RuntimeError("foo")

    payload = unhexlify("A2616103616205")
    with pytest.raises(CBORDecodeError) as exc_info:
        loads(payload, object_hook=object_hook)

    assert isinstance(exc_info.value.__cause__, RuntimeError)
    assert exc_info.value.__cause__.args[0] == "foo"


class TestCustomDecoder:
    def test_success(self) -> None:
        def custom_set_decoder(value: Any, immutable: bool) -> Any:
            return frozenset(value) if immutable else set(value)

        payload = unhexlify("d9010283050763666f6f")
        assert loads(payload, semantic_decoders={258: custom_set_decoder}) == {5, 7, "foo"}
        assert loads(
            payload, semantic_decoders={258: custom_set_decoder}, immutable=True
        ) == frozenset((5, 7, "foo"))

    def test_cyclic_reference(self) -> None:
        @dataclass
        class CustomContainer:
            value: Any

        @shareable_decoder(name="custom container")
        def custom_decoder(
            immutable: bool,
        ) -> tuple[CustomContainer | None, Callable[[Any], Any]]:
            container = None if immutable else CustomContainer(None)

            def callback(item: Any) -> CustomContainer:
                if container is None:
                    return CustomContainer(item)
                else:
                    container.value = item
                    return container

            return container, callback

        payload = unhexlify("d81cda00013880d81d00")
        retval = loads(payload, semantic_decoders={80_000: custom_decoder})
        assert retval.value is retval

    @pytest.mark.parametrize(
        "immutable, expected_value",
        [
            pytest.param(False, [1, 2], id="mutable"),
            pytest.param(True, (1, 2), id="immutable"),
        ],
    )
    def test_immutable_flag(self, immutable: bool, expected_value: Any) -> None:
        @dataclass
        class CustomContainer:
            value: tuple[Any, ...] | list[Any]

        @shareable_decoder(name="custom container", immutable=immutable)
        def custom_decoder(
            immutable: bool,
        ) -> tuple[CustomContainer | None, Callable[[Any], Any]]:
            def callback(item: Any) -> CustomContainer:
                return CustomContainer(item)

            return None, callback

        payload = unhexlify("da00013880820102")
        retval = loads(payload, semantic_decoders={80_000: custom_decoder})
        assert isinstance(retval, CustomContainer)
        assert retval.value == expected_value

    @pytest.mark.parametrize(
        "name, expected_typename",
        [
            pytest.param(None, "semantic tag 80000"),
            pytest.param("custom container", "custom container"),
        ],
    )
    def test_error_in_callback(self, name: str | None, expected_typename: str) -> None:
        @shareable_decoder(name=name)
        def custom_decoder(immutable: bool) -> tuple[None, Callable[[Any], Any]]:
            def callback(item: Any) -> NoReturn:
                raise RuntimeError("foo")

            return None, callback

        payload = unhexlify("da000138806474657374")
        with pytest.raises(
            CBORDecodeError, match=f"error decoding {expected_typename}"
        ) as exc_info:
            loads(payload, semantic_decoders={80_000: custom_decoder})

        assert isinstance(exc_info.value.__cause__, RuntimeError)
        assert str(exc_info.value.__cause__) == "foo"


def test_load_from_file(tmp_path: Path) -> None:
    path = tmp_path / "testdata.cbor"
    path.write_bytes(b"\x82\x01\x0a")
    with path.open("rb") as fp:
        obj = load(fp)

    assert obj == [1, 10]


def test_nested_dict() -> None:
    value = loads(unhexlify("A1D9177082010201"))
    assert type(value) is dict
    assert value == {CBORTag(6000, (1, 2)): 1}


def test_set() -> None:
    payload = unhexlify("d9010283616361626161")
    value = loads(payload)
    assert type(value) is set
    assert value == {"a", "b", "c"}


@pytest.mark.parametrize(
    "payload, expected",
    [
        ("a1a1616161626163", {frozendict({"a": "b"}): "c"}),
        (
            "a1a1a10101a1666e6573746564f5a1666e6573746564f4",
            {frozendict({frozendict({1: 1}): frozendict({"nested": True})}): {"nested": False}},
        ),
        ("a182010203", {(1, 2): 3}),
        ("a1d901028301020304", {frozenset({1, 2, 3}): 4}),
        ("a17f657374726561646d696e67ff01", {"streaming": 1}),
        ("d9010282d90102820102d90102820304", {frozenset({1, 2}), frozenset({3, 4})}),
    ],
)
def test_immutable_keys(payload: str, expected: object) -> None:
    value = loads(unhexlify(payload))
    assert value == expected


# Corrupted or invalid data checks


def test_huge_truncated_array(will_overflow: bytes) -> None:
    with pytest.raises(CBORDecodeError):
        loads(unhexlify("9b") + will_overflow)


def test_huge_truncated_string() -> None:
    huge_index = struct.pack("Q", sys.maxsize + 1)
    with pytest.raises((CBORDecodeError, MemoryError)):
        loads(unhexlify("7b") + huge_index + unhexlify("70717273"))


@pytest.mark.parametrize(
    "dtype_prefix", [pytest.param("7B", id="string"), pytest.param("5b", id="bytes")]
)
def test_huge_truncated_data(dtype_prefix: str, will_overflow: bytes) -> None:
    with pytest.raises((CBORDecodeError, MemoryError)):
        loads(unhexlify(dtype_prefix) + will_overflow)


@pytest.mark.parametrize(
    "tag_dtype", [pytest.param("7F7B", id="string"), pytest.param("5f5B", id="bytes")]
)
def test_huge_truncated_indefinite_data(tag_dtype: str, will_overflow: bytes) -> None:
    huge_index = struct.pack("Q", sys.maxsize + 1)
    with pytest.raises((CBORDecodeError, MemoryError)):
        loads(unhexlify(tag_dtype) + huge_index + unhexlify("70717273ff"))


@pytest.mark.parametrize(
    "data",
    [
        pytest.param("7f61777f6177ffff", id="string"),
        pytest.param("5f41775f4177ffff", id="bytes"),
    ],
)
def test_embedded_indefinite_data(data: str) -> None:
    with pytest.raises(CBORDecodeError):
        loads(unhexlify(data))


@pytest.mark.parametrize(
    "data", [pytest.param("7f01ff", id="string"), pytest.param("5f01ff", id="bytes")]
)
def test_invalid_indefinite_data_item(data: str) -> None:
    with pytest.raises(CBORDecodeError):
        loads(unhexlify(data))


@pytest.mark.parametrize(
    "data",
    [
        pytest.param("7f7bff0000000000000471717272ff", id="string"),
        pytest.param("5f5bff0000000000000471717272ff", id="bytes"),
    ],
)
def test_indefinite_overflow(data: str) -> None:
    with pytest.raises(CBORDecodeError):
        loads(unhexlify(data))


def test_invalid_cbor() -> None:
    with pytest.raises(CBORDecodeError):
        loads(
            unhexlify(
                "c788370016b8965bdb2074bff82e5a20e09bec21f8406e86442b87ec3ff245b70a47624dc9cdc682"
                "4b2a4c52e95ec9d6b0534b71c2b49e4bf9031500cee6869979c297bb5a8b381e98db714108415e5c"
                "50db78974c271579b01633a3ef6271be5c225eb2"
            )
        )


@pytest.mark.parametrize(
    "data, expected",
    [("fc", "1c"), ("fd", "1d"), ("fe", "1e")],
)
def test_reserved_special_tags(data: str, expected: str) -> None:
    with pytest.raises(
        CBORDecodeError, match=f"undefined reserved major type 7 subtype 0x{expected}"
    ):
        loads(unhexlify(data))


@pytest.mark.parametrize(
    "data, expected, typename",
    [("c400", "4", "decimal fraction"), ("c500", "5", "bigfloat")],
)
def test_decimal_payload_unpacking(data: str, expected: str, typename: str) -> None:
    with pytest.raises(CBORDecodeError, match=f"error decoding {typename}") as exc_info:
        loads(unhexlify(data))

    assert isinstance(exc_info.value.__cause__, TypeError)
    assert str(exc_info.value.__cause__) == "input value must be an array"


@pytest.mark.parametrize(
    "payload",
    [
        pytest.param(
            unhexlify("41"),
            id="bytestring",
        ),
        pytest.param(
            unhexlify("61"),
            id="unicode",
        ),
    ],
)
def test_oversized_read(payload: bytes, tmp_path: Path) -> None:
    with pytest.raises(CBORDecodeEOF, match="premature end of stream"):
        dummy_path = tmp_path / "testdata"
        dummy_path.write_bytes(payload)
        with dummy_path.open("rb") as f:
            load(f)


class TestDecoderReuse:
    """
    Tests for correct behavior when reusing CBORDecoder instances.
    """

    def test_decoder_reuse_resets_shared_refs(self) -> None:
        """
        Shared references should be scoped to a single decode operation,
        not persist across multiple decodes on the same decoder instance.
        """
        # Message with shareable tag (28)
        msg1 = dumps(CBORTag(28, "first_value"))

        # Message with sharedref tag (29) referencing index 0
        msg2 = dumps(CBORTag(29, 0))

        # Reuse decoder across messages
        decoder = CBORDecoder(BytesIO(msg1))
        result1 = decoder.decode()
        assert result1 == "first_value"

        # Second decode should fail - sharedref(0) doesn't exist in this context
        decoder.fp = BytesIO(msg2)
        with pytest.raises(CBORDecodeError, match="shared reference"):
            decoder.decode()

    def test_shared_refs_within_single_decode(self) -> None:
        """
        Shared references must work correctly within a single decode operation.

        Note: This tests non-cyclic sibling references [shareable(x), sharedref(0)],
        which is a different pattern from test_cyclic_array/test_cyclic_map that
        test self-referencing structures like shareable([sharedref(0)]).
        """
        # [shareable("hello"), sharedref(0)] -> ["hello", "hello"]
        data = unhexlify(
            "82"  # array(2)
            "d81c"  # tag(28) shareable
            "65"  # text(5)
            "68656c6c6f"  # "hello"
            "d81d"  # tag(29) sharedref
            "00"  # unsigned(0)
        )

        result = loads(data)
        assert result == ["hello", "hello"]
        assert result[0] is result[1]  # Same object reference


def test_str_errors_valid_utf8_unchanged() -> None:
    payload = b"\x78\x19Hello \xc3\xbcnicode \xe6\xb0\xb4 world!"
    result_strict = loads(payload, str_errors="strict")
    result_replace = loads(payload, str_errors="replace")
    assert result_strict == result_replace
    assert result_strict == "Hello \u00fcnicode \u6c34 world!"


@pytest.mark.parametrize("length", [255, 256, 257])
def test_string_stack_threshold_boundary(length: int) -> None:
    """Test stack (<=256) vs heap (>256) allocation boundary."""
    test_string = "a" * length
    if length < 24:
        payload = bytes([0x60 + length])
    elif length < 256:
        payload = b"\x78" + bytes([length])
    else:
        payload = b"\x79" + struct.pack(">H", length)

    payload += test_string.encode("utf-8")
    assert loads(payload) == test_string


def test_override_semantic_decoder() -> None:
    def date_decoder(value: int, immutable: bool) -> datetime:
        assert value == 1363896240
        return datetime(2026, 2, 18)

    payload = unhexlify("c11a514b67b0")  # datetime(2013, 3, 21, 20, 4, 0, tzinfo=timezone.utc)
    assert loads(payload, semantic_decoders={1: date_decoder}) == datetime(2026, 2, 18)


@pytest.mark.parametrize(
    "input_type",
    [
        pytest.param(bytes, id="bytes"),
        pytest.param(bytearray, id="bytearray"),
        pytest.param(memoryview, id="memoryview"),
    ],
)
def test_loads_buffer_input(input_type: type) -> None:
    payload = unhexlify("82010a")
    assert loads(input_type(payload)) == [1, 10]


@pytest.mark.parametrize(
    "data",
    [
        pytest.param([b"x" * 4096], id="read"),
        pytest.param([b"x" * 4089, [b"x" * 4096]], id="read_exact"),
    ],
)
def test_load_exceeds_buffer_size(data: object) -> None:
    """Regression test for #304."""
    payload = dumps(data)
    buf = BytesIO(payload)
    assert load(buf) == data


def test_load_buffer_truncation() -> None:
    """
    With read_size=4096, load() will request more bytes than are currently
    available in the internal buffer (requesting 8 bytes while only 6 remain),
    triggering buffer truncation/extension.
    """
    data = [3.14] * 455
    payload = dumps(data)
    buf = BytesIO(payload)
    assert load(buf, read_size=4096) == data



import sys
from collections.abc import Callable, Iterator, Mapping, Sequence
from datetime import date, datetime, tzinfo
from decimal import Decimal
from email.mime.text import MIMEText
from fractions import Fraction
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from re import Pattern
from typing import IO, Any, Generic, TypeAlias, final, overload
from uuid import UUID

if sys.hexversion < 51314855:
    @final
    class frozendict(Generic[_KT, _VT_co], Mapping[_KT, _VT_co]):
        def __new__(cls, *args: Any, **kwargs: Any) -> Self: ...
        def __getitem__(self, key: _KT, /) -> _VT_co: ...
        def __len__(self) -> int: ...
        def __iter__(self) -> Iterator[_T_co]: ...

if sys.version_info >= (3, 13):
    from typing import TypeVar
else:
    from typing_extensions import TypeVar

if sys.version_info >= (3, 12):
    from collections.abc import Buffer
else:
    from typing_extensions import Buffer

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_KT = TypeVar("_KT")
_VT_co = TypeVar("_VT_co", covariant=True)
_TContainer = TypeVar("_TContainer", default=None)

TagHook: TypeAlias = Callable[[CBORTag, bool], Any]
SemanticDecoderCallback: TypeAlias = Callable[[Any, bool], Any]
ObjectHook: TypeAlias = Callable[[Mapping[Any, Any], bool], Any]
EncoderHook: TypeAlias = Callable[[CBOREncoder, Any], Any]
ShareableDecoderCallback: TypeAlias = Callable[[Any], Any]
ShareableDecoderInitializer: TypeAlias = Callable[[bool], tuple[Any, ShareableDecoderCallback]]

@final
class CBOREncoder:
    datetime_as_timestamp: bool
    timezone: tzinfo | None
    value_sharing: bool
    default: EncoderHook | None
    canonical: bool
    date_as_datetime: bool
    string_referencing: bool
    indefinite_containers: bool
    def __new__(
        cls,
        fp: IO[bytes],
        *,
        datetime_as_timestamp: bool = ...,
        timezone: tzinfo | None = ...,
        value_sharing: bool = ...,
        encoders: Mapping[type, EncoderHook] | None = ...,
        default: EncoderHook | None = ...,
        canonical: bool = ...,
        date_as_datetime: bool = ...,
        string_referencing: bool = ...,
        indefinite_containers: bool = ...,
    ) -> Self: ...
    fp: IO[bytes]
    def flush(self) -> None: ...
    def write(self, buf: bytes, /) -> int: ...
    def encode(self, obj: object, /) -> None: ...
    def encode_to_bytes(self, obj: object, /) -> bytes: ...
    def encode_length(self, major_tag: int, length: int | None) -> None: ...
    def encode_string(self, obj: str) -> None: ...
    def encode_bytes(self, obj: bytes) -> None: ...
    def encode_bytearray(self, obj: bytearray) -> None: ...
    def encode_array(self, obj: Sequence[Any]) -> None: ...
    def encode_map(self, obj: Mapping[Any, Any]) -> None: ...
    def encode_break(self) -> None: ...
    def encode_int(self, obj: int) -> None: ...
    def encode_bool(self, obj: bool) -> None: ...
    def encode_none(self) -> None: ...
    def encode_undefined(self) -> None: ...
    def encode_semantic(self, tag: int, value: Any) -> None: ...
    def encode_set(self, obj: set[Any]) -> None: ...
    def encode_frozenset(self, obj: frozenset[Any]) -> None: ...
    def encode_datetime(self, obj: datetime) -> None: ...
    def encode_date(self, obj: date) -> None: ...
    def encode_rational(self, obj: Fraction) -> None: ...
    def encode_regexp(self, obj: Pattern[str]) -> None: ...
    def encode_mime(self, obj: MIMEText) -> None: ...
    def encode_uuid(self, obj: UUID) -> None: ...
    def encode_decimal(self, obj: Decimal) -> None: ...
    def encode_ipv4_address(self, obj: IPv4Address) -> None: ...
    def encode_ipv4_network(self, obj: IPv4Network) -> None: ...
    def encode_ipv4_interface(self, obj: IPv4Interface) -> None: ...
    def encode_ipv6_address(self, obj: IPv6Address) -> None: ...
    def encode_ipv6_network(self, obj: IPv6Network) -> None: ...
    def encode_ipv6_interface(self, obj: IPv6Interface) -> None: ...
    def encode_simple_value(self, obj: CBORSimpleValue) -> None: ...
    def encode_float(self, obj: float) -> None: ...
    def encode_complex(self, obj: complex) -> None: ...

@final
class CBORDecoder:
    fp: IO[bytes]
    tag_hook: TagHook | None
    object_hook: ObjectHook | None
    str_errors: str
    def __new__(
        cls,
        fp: IO[bytes],
        *,
        tag_hook: TagHook | None = ...,
        object_hook: ObjectHook | None = ...,
        semantic_decoders: Mapping[int, SemanticDecoderCallback | ShareableDecoderInitializer]
        | None = ...,
        str_errors: str = ...,
        read_size: int = ...,
        max_depth: int = ...,
        allow_indefinite: bool = ...,
        allow_duplicate_keys: bool = ...,
    ) -> Self: ...

    # Properties
    @property
    def read_size(self) -> int: ...
    @property
    def max_depth(self) -> int: ...
    @property
    def allow_indefinite(self) -> bool: ...
    @property
    def allow_duplicate_keys(self) -> bool: ...
    def decode(self, *, immutable: bool = ...) -> Any: ...
    def read(self, amount: int, /) -> bytes: ...

class CBORError(Exception): ...
class CBOREncodeError(CBORError): ...
class CBOREncodeTypeError(CBOREncodeError, TypeError): ...
class CBOREncodeValueError(CBOREncodeError, ValueError): ...
class CBORDecodeError(CBORError): ...
class CBORDecodeEOF(CBORDecodeError, EOFError): ...

@final
class CBORTag:
    tag: int
    value: Any

    def __new__(cls, tag: int, value: Any) -> Self: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object, /) -> bool: ...
    def __gt__(self, other: object, /) -> bool: ...
    def __ge__(self, other: object, /) -> bool: ...

@final
class CBORSimpleValue:
    value: int

    def __new__(cls, value: int) -> Self: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object, /) -> bool: ...
    def __gt__(self, other: object, /) -> bool: ...
    def __ge__(self, other: object, /) -> bool: ...

@final
class UndefinedType: ...

undefined: UndefinedType

def dump(
    obj: object,
    fp: IO[bytes],
    *,
    datetime_as_timestamp: bool = ...,
    timezone: tzinfo | None = ...,
    value_sharing: bool = ...,
    encoders: Mapping[type, EncoderHook] | None = ...,
    default: EncoderHook | None = ...,
    canonical: bool = ...,
    date_as_datetime: bool = ...,
    string_referencing: bool = ...,
    indefinite_containers: bool = ...,
) -> None: ...
def dumps(
    obj: object,
    *,
    datetime_as_timestamp: bool = ...,
    timezone: tzinfo | None = ...,
    value_sharing: bool = ...,
    encoders: Mapping[type, EncoderHook] | None = ...,
    default: EncoderHook | None = ...,
    canonical: bool = ...,
    date_as_datetime: bool = ...,
    string_referencing: bool = ...,
    indefinite_containers: bool = ...,
) -> bytes: ...
def load(
    fp: IO[bytes],
    *,
    tag_hook: TagHook | None = ...,
    object_hook: ObjectHook | None = ...,
    semantic_decoders: Mapping[int, SemanticDecoderCallback | ShareableDecoderInitializer]
    | None = ...,
    str_errors: str = ...,
    read_size: int = ...,
    max_depth: int = ...,
    allow_indefinite: bool = ...,
    allow_duplicate_keys: bool = ...,
    immutable: bool = ...,
) -> Any: ...
def loads(
    data: Buffer,
    *,
    tag_hook: TagHook | None = ...,
    object_hook: ObjectHook | None = ...,
    semantic_decoders: Mapping[int, SemanticDecoderCallback | ShareableDecoderInitializer]
    | None = ...,
    str_errors: str = ...,
    max_depth: int = ...,
    allow_indefinite: bool = ...,
    allow_duplicate_keys: bool = ...,
    immutable: bool = ...,
) -> Any: ...
def shareable_encoder(
    wraps: Callable[[CBOREncoder, _T], None], /
) -> Callable[[CBOREncoder, _T], None]: ...
@overload
def shareable_decoder(func: ShareableDecoderInitializer, /) -> ShareableDecoderInitializer: ...
@overload
def shareable_decoder(
    *, name: str | None = ..., immutable: bool = ...
) -> Callable[[ShareableDecoderInitializer], ShareableDecoderCallback]: ...



import re
import sys
from binascii import unhexlify
from collections import OrderedDict
from collections.abc import Set
from datetime import date, datetime, timedelta, timezone
from decimal import Decimal
from email.mime.text import MIMEText
from fractions import Fraction
from io import BytesIO
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from pathlib import Path
from typing import Any
from uuid import UUID

import pytest
from _pytest.fixtures import FixtureRequest, SubRequest
from hypothesis import given

from cbor2 import (
    CBOREncodeError,
    CBOREncoder,
    CBORSimpleValue,
    CBORTag,
    dump,
    dumps,
    loads,
    shareable_encoder,
    undefined,
)

from .hypothesis_strategies import compound_types_strategy

if sys.hexversion < 51314855:
    from cbor2 import frozendict


class TestFpAttribute:
    def test_none(self) -> None:
        with pytest.raises(ValueError, match=r"fp must be a writable file-like object"):
            CBOREncoder(None)  # type: ignore[arg-type]

    def test_not_writable(self, tmp_path: Path) -> None:
        # Test for fp not being writable
        path = tmp_path.joinpath("foo.cbor")
        path.touch()
        with (
            pytest.raises(ValueError, match=r"fp must be a writable file-like object"),
            path.open("rb") as fp,
        ):
            CBOREncoder(fp)

    def test_delete(self) -> None:
        decoder = CBOREncoder(BytesIO())
        with pytest.raises(AttributeError):
            del decoder.fp


def test_default_attr() -> None:
    with BytesIO() as stream:
        encoder = CBOREncoder(stream)
        assert encoder.default is None
        with pytest.raises(TypeError):
            encoder.default = 1  # type: ignore[assignment]
        with pytest.raises(AttributeError):
            del encoder.default


def test_timezone_attr() -> None:
    with BytesIO() as stream:
        encoder = CBOREncoder(stream)
        assert encoder.timezone is None
        with pytest.raises(TypeError):
            encoder.timezone = 1  # type: ignore[assignment]
        with pytest.raises(AttributeError):
            del encoder.timezone


def test_write() -> None:
    with BytesIO() as stream:
        encoder = CBOREncoder(stream)
        encoder.write(b"foo")
        assert stream.getvalue() == b"foo"
        with pytest.raises(TypeError):
            encoder.write(1)  # type: ignore[arg-type]


def test_encode_length() -> None:
    fp = BytesIO()
    encoder = CBOREncoder(fp)

    def reset_encoder() -> None:
        nonlocal fp, encoder
        fp = BytesIO()
        encoder = CBOREncoder(fp)

    encoder.encode_length(0, 1)
    assert fp.getvalue() == b"\x01"

    # Array of size 2
    reset_encoder()
    encoder.encode_length(4, 2)
    assert fp.getvalue() == b"\x82"

    # Array of indefinite size
    reset_encoder()
    encoder.encode_length(4, None)
    assert fp.getvalue() == b"\x9f"

    # Map of size 0
    reset_encoder()
    encoder.encode_length(5, 0)
    assert fp.getvalue() == b"\xa0"

    # Map of indefinite size
    reset_encoder()
    encoder.encode_length(5, None)
    assert fp.getvalue() == b"\xbf"

    # Indefinite container break
    reset_encoder()
    encoder.encode_break()
    assert fp.getvalue() == b"\xff"


def test_canonical_attr() -> None:
    # Another test purely for coverage in the C variant
    with BytesIO() as stream:
        enc = CBOREncoder(stream)
        assert not enc.canonical
        enc = CBOREncoder(stream, canonical=True)
        assert enc.canonical


@pytest.mark.parametrize(
    "value, expected",
    [
        (0, "00"),
        (1, "01"),
        (10, "0a"),
        (23, "17"),
        (24, "1818"),
        (100, "1864"),
        (1000, "1903e8"),
        (1000000, "1a000f4240"),
        (1000000000000, "1b000000e8d4a51000"),
        (18446744073709551615, "1bffffffffffffffff"),
        (18446744073709551616, "c249010000000000000000"),
        (-18446744073709551616, "3bffffffffffffffff"),
        (-18446744073709551617, "c349010000000000000000"),
        (-1, "20"),
        (-10, "29"),
        (-100, "3863"),
        (-1000, "3903e7"),
    ],
)
def test_integer(value: int, expected: str) -> None:
    assert dumps(value) == unhexlify(expected)


@pytest.mark.parametrize(
    "value, expected",
    [
        (1.1, "fb3ff199999999999a"),
        (1.0e300, "fb7e37e43c8800759c"),
        (-4.1, "fbc010666666666666"),
        (float("inf"), "f97c00"),
        (float("nan"), "f97e00"),
        (float("-inf"), "f9fc00"),
    ],
)
def test_float(value: float, expected: str) -> None:
    assert dumps(value) == unhexlify(expected)


@pytest.mark.parametrize(
    "value, expected",
    [
        (b"", "40"),
        (b"\x01\x02\x03\x04", "4401020304"),
    ],
)
def test_bytestring(value: bytes, expected: str) -> None:
    assert dumps(value) == unhexlify(expected)


def test_bytearray() -> None:
    expected = unhexlify("4401020304")
    assert dumps(bytearray(b"\x01\x02\x03\x04")) == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        ("", "60"),
        ("a", "6161"),
        ("IETF", "6449455446"),
        ('"\\', "62225c"),
        ("\u00fc", "62c3bc"),
        ("\u6c34", "63e6b0b4"),
    ],
)
def test_string(value: str, expected: str) -> None:
    assert dumps(value) == unhexlify(expected)


@pytest.fixture(
    params=[
        pytest.param((False, "f4"), id="false"),
        pytest.param((True, "f5"), id="true"),
        pytest.param((None, "f6"), id="null"),
        pytest.param((undefined, "f7"), id="undefined"),
    ]
)
def special_values(request: FixtureRequest) -> tuple[object, str]:
    value, expected = request.param
    if value == "undefined":
        value = undefined

    return value, expected


def test_special(special_values: tuple[object, str]) -> None:
    value, expected = special_values
    assert dumps(value) == unhexlify(expected)


@pytest.fixture(params=[(0, "e0"), (2, "e2"), (23, "f7"), (32, "f820")])
def simple_values(request: SubRequest) -> tuple[CBORSimpleValue, str]:
    value, expected = request.param
    return CBORSimpleValue(value), expected


def test_simple_value(simple_values: tuple[CBORSimpleValue, str]) -> None:
    value, expected = simple_values
    assert dumps(value) == unhexlify(expected)


def test_simple_val_as_key() -> None:
    payload = {CBORSimpleValue(99): 1}
    result = dumps(payload)
    assert result == unhexlify("A1F86301")


#
# Tests for extension tags
#


@pytest.mark.parametrize(
    "value, as_timestamp, expected",
    [
        pytest.param(
            datetime(2013, 3, 21, 20, 4, 0, tzinfo=timezone.utc),
            False,
            "c074323031332d30332d32315432303a30343a30305a",
            id="datetime/utc",
        ),
        pytest.param(
            datetime(2013, 3, 21, 20, 4, 0, 380841, tzinfo=timezone.utc),
            False,
            "c0781b323031332d30332d32315432303a30343a30302e3338303834315a",
            id="datetime+micro/utc",
        ),
        pytest.param(
            datetime(2013, 3, 21, 22, 4, 0, tzinfo=timezone(timedelta(hours=2))),
            False,
            "c07819323031332d30332d32315432323a30343a30302b30323a3030",
            id="datetime/eet",
        ),
        pytest.param(
            datetime(2013, 3, 21, 20, 4, 0),
            False,
            "c074323031332d30332d32315432303a30343a30305a",
            id="naive",
        ),
        pytest.param(
            datetime(2013, 3, 21, 20, 4, 0, tzinfo=timezone.utc),
            True,
            "c11a514b67b0",
            id="timestamp/utc",
        ),
        pytest.param(
            datetime(2013, 3, 21, 20, 4, 0, 123456, tzinfo=timezone.utc),
            True,
            "c1fb41d452d9ec07e6b4",
            id="timestamp+micro/utc",
        ),
        pytest.param(
            datetime(2013, 3, 21, 22, 4, 0, tzinfo=timezone(timedelta(hours=2))),
            True,
            "c11a514b67b0",
            id="timestamp/eet",
        ),
    ],
)
def test_datetime(value: datetime, as_timestamp: bool, expected: str) -> None:
    assert dumps(value, datetime_as_timestamp=as_timestamp, timezone=timezone.utc) == unhexlify(
        expected
    )


@pytest.mark.parametrize(
    "value, as_timestamp, expected",
    [
        pytest.param(
            date(2013, 3, 21),
            False,
            "d903ec6a323031332d30332d3231",
            id="date/string",
        ),
        pytest.param(
            date(2018, 12, 31),
            True,
            "d8641945e8",
            id="date/timestamp",
        ),
    ],
)
def test_date(value: date, as_timestamp: bool, expected: str) -> None:
    assert dumps(value, datetime_as_timestamp=as_timestamp) == unhexlify(expected)


def test_date_as_datetime() -> None:
    expected = unhexlify("c074323031332d30332d32315430303a30303a30305a")
    assert dumps(date(2013, 3, 21), timezone=timezone.utc, date_as_datetime=True) == expected


def test_naive_datetime() -> None:
    """Test that naive datetimes are gracefully rejected when no timezone has been set."""
    with pytest.raises(CBOREncodeError) as exc:
        dumps(datetime(2013, 3, 21))
        exc.match(
            "naive datetime datetime.datetime(2013, 3, 21) encountered "
            "and no default timezone has been set"
        )
        assert isinstance(exc, ValueError)


@pytest.mark.parametrize(
    "value, expected",
    [
        pytest.param(Decimal("14.123"), "c4822219372b", id="normal"),
        pytest.param(Decimal("-14.123"), "C4822239372A", id="negative"),
        pytest.param(Decimal("NaN"), "f97e00", id="nan"),
        pytest.param(Decimal("Infinity"), "f97c00", id="inf"),
        pytest.param(Decimal("-Infinity"), "f9fc00", id="neginf"),
    ],
)
def test_decimal(value: Decimal, expected: str) -> None:
    assert dumps(value) == unhexlify(expected)


@pytest.mark.parametrize(
    "value, expected",
    [
        (3.1 + 2.1j, "d9a7f882fb4008cccccccccccdfb4000cccccccccccd"),
        (1.0e300j, "d9a7f882fb0000000000000000fb7e37e43c8800759c"),
        (0.0j, "d9a7f882fb0000000000000000fb0000000000000000"),
        (complex(float("inf"), float("inf")), "d9a7f882f97c00f97c00"),
        (complex(float("inf"), 0.0), "d9a7f882f97c00fb0000000000000000"),
        (complex(float("nan"), float("inf")), "d9a7f882f97e00f97c00"),
    ],
)
def test_complex(value: complex, expected: str) -> None:
    assert dumps(value) == unhexlify(expected)


def test_rational() -> None:
    assert dumps(Fraction(2, 5)) == unhexlify("d81e820205")


def test_regex() -> None:
    expected = unhexlify("d8236d68656c6c6f2028776f726c6429")
    assert dumps(re.compile("hello (world)")) == expected


def test_mime() -> None:
    expected = unhexlify(
        "d824787b436f6e74656e742d547970653a20746578742f706c61696e3b20636861727365743d2269736f2d38"
        "3835392d3135220a4d494d452d56657273696f6e3a20312e300a436f6e74656e742d5472616e736665722d456"
        "e636f64696e673a2071756f7465642d7072696e7461626c650a0a48656c6c6f203d413475726f"
    )
    message = MIMEText("Hello \u20acuro", "plain", "iso-8859-15")
    assert dumps(message) == expected


def test_uuid() -> None:
    expected = unhexlify("d825505eaffac8b51e480581277fdcc7842faf")
    assert dumps(UUID(hex="5eaffac8b51e480581277fdcc7842faf")) == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        pytest.param(IPv4Address("192.0.2.1"), "d83444c0000201", id="ipv4addr"),
        pytest.param(IPv4Network("192.0.2.0/24"), "d83482181843c00002", id="ipv4net"),
        pytest.param(IPv4Interface("192.0.2.1/24"), "d8348244c00002011818", id="ipv4if"),
        pytest.param(
            IPv6Address("2001:0db8:1234:deed:beef:cafe:face:feed"),
            "d8365020010db81234deedbeefcafefacefeed",
            id="ipv6addr",
        ),
        pytest.param(
            IPv6Network("2001:db8:1234::/48"),
            "d8368218304620010db81234",
            id="ipv6net",
        ),
        pytest.param(
            IPv6Interface("fe80::202:2ff:ffff:fe03:303%eth0/64"),
            "d8368350fe8000000000020202fffffffe03030318404465746830",
            id="ipv6if",
        ),
    ],
)
def test_ipaddress(value: object, expected: str) -> None:
    assert dumps(value) == unhexlify(expected)


def test_custom_tag() -> None:
    expected = unhexlify("d917706548656c6c6f")
    assert dumps(CBORTag(6000, "Hello")) == expected


def test_cyclic_array() -> None:
    """Test that an array that contains itself can be serialized with value sharing enabled."""
    expected = unhexlify("d81c81d81c81d81d00")
    a: list[list[Any]] = [[]]
    a[0].append(a)
    assert dumps(a, value_sharing=True) == expected


def test_cyclic_array_nosharing() -> None:
    """Test that serializing a cyclic structure w/o value sharing will blow up gracefully."""
    a: list[list[Any]] = []
    a.append(a)
    with pytest.raises(CBOREncodeError) as exc:
        dumps(a)
        exc.match("cyclic data structure detected but value sharing is disabled")
        assert isinstance(exc, ValueError)


def test_cyclic_map() -> None:
    """Test that a dict that contains itself can be serialized with value sharing enabled."""
    expected = unhexlify("d81ca100d81d00")
    a: dict[int, dict[int, Any]] = {}
    a[0] = a
    assert dumps(a, value_sharing=True) == expected


def test_cyclic_map_nosharing() -> None:
    """Test that serializing a cyclic structure w/o value sharing will fail gracefully."""
    a: dict[int, dict[int, Any]] = {}
    a[0] = a
    with pytest.raises(CBOREncodeError) as exc:
        dumps(a)
        exc.match("cyclic data structure detected but value sharing is disabled")
        assert isinstance(exc, ValueError)


@pytest.mark.parametrize(
    "value_sharing, expected",
    [
        pytest.param(False, "828080", id="nosharing"),
        pytest.param(True, "d81c82d81c80d81d01", id="sharing"),
    ],
)
def test_not_cyclic_same_object(value_sharing: bool, expected: str) -> None:
    """Test that the same shareable object can be included twice if not in a cyclic structure."""
    a: list[Any] = []
    b = [a, a]
    assert dumps(b, value_sharing=value_sharing) == unhexlify(expected)


def test_unsupported_type() -> None:
    with pytest.raises(CBOREncodeError) as exc:
        dumps(lambda: None)
        exc.match("cannot serialize type function")
        assert isinstance(exc, TypeError)


def test_default() -> None:
    class DummyType:
        def __init__(self, state: object):
            self.state = state

    def default_encoder(encoder: CBOREncoder, value: Any) -> None:
        encoder.encode(value.state)

    expected = unhexlify("820305")
    obj = DummyType([3, 5])
    serialized = dumps(obj, default=default_encoder)
    assert serialized == expected


def test_default_cyclic() -> None:
    class DummyType:
        def __init__(self, value: object = None):
            self.value = value

    @shareable_encoder
    def default_encoder(encoder: CBOREncoder, value: Any) -> None:
        state = encoder.encode_to_bytes(value.value)
        encoder.encode(CBORTag(3000, state))

    expected = unhexlify("D81CD90BB849D81CD90BB843D81D00")
    obj = DummyType()
    obj2 = DummyType(obj)
    obj.value = obj2
    serialized = dumps(obj, value_sharing=True, default=default_encoder)
    assert serialized == expected


def test_dump_to_file(tmp_path: Path) -> None:
    path = tmp_path / "testdata.cbor"
    with path.open("wb") as fp:
        dump([1, 10], fp)

    assert path.read_bytes() == b"\x82\x01\x0a"


@pytest.mark.parametrize(
    "value, expected",
    [
        pytest.param({}, "a0", id="empty"),
        pytest.param(
            OrderedDict([(b"a", b""), (b"b", b"")]), "A2416140416240", id="bytes in order"
        ),
        pytest.param(
            OrderedDict([(b"b", b""), (b"a", b"")]), "A2416140416240", id="bytes out of order"
        ),
        pytest.param(OrderedDict([("a", ""), ("b", "")]), "a2616160616260", id="text in order"),
        pytest.param(
            OrderedDict([("b", ""), ("a", "")]), "a2616160616260", id="text out of order"
        ),
        pytest.param(
            OrderedDict([(b"00001", ""), (b"002", "")]),
            "A2433030326045303030303160",
            id="byte length",
        ),
        pytest.param(OrderedDict([(255, 0), (2, 0)]), "a2020018ff00", id="integer keys"),
        pytest.param(frozendict([(b"a", b""), (b"b", b"")]), "A2416140416240", id="frozendict"),
    ],
)
def test_ordered_map(value: object, expected: str) -> None:
    assert dumps(value, canonical=True) == unhexlify(expected)


@pytest.mark.parametrize(
    "value, expected",
    [
        pytest.param(3.5, "F94300", id="float 16"),
        pytest.param(100000.0, "FA47C35000", id="float 32"),
        pytest.param(3.8, "FB400E666666666666", id="float 64"),
        pytest.param(float("inf"), "f97c00", id="inf"),
        pytest.param(float("nan"), "f97e00", id="nan"),
        pytest.param(float("-inf"), "f9fc00", id="-inf"),
        pytest.param(
            float.fromhex("0x1.0p-24"),
            "f90001",
            id="float 16 minimum positive subnormal",
        ),
        pytest.param(
            float.fromhex("0x1.4p-24"),
            "fa33a00000",
            id="mantissa o/f to 32",
        ),
        pytest.param(
            float.fromhex("0x1.ff8p-63"),
            "fa207fc000",
            id="exponent o/f to 32",
        ),
        pytest.param(1e300, "fb7e37e43c8800759c", id="oversize float"),
    ],
)
def test_minimal_floats(value: float, expected: str) -> None:
    assert dumps(value, canonical=True) == unhexlify(expected)


def test_tuple_key() -> None:
    assert dumps({(2, 1): ""}) == unhexlify("a182020160")


def test_dict_key() -> None:
    assert dumps({frozendict({2: 1}): ""}) == unhexlify("a1a1020160")


@pytest.mark.parametrize(
    "frozen",
    [
        pytest.param(False, id="set"),
        pytest.param(True, id="frozenset"),
    ],
)
def test_set(frozen: bool) -> None:
    value: Set[str] = {"a", "b", "c"}
    if frozen:
        value = frozenset(value)

    serialized = dumps(value)
    assert len(serialized) == 10
    assert serialized.startswith(unhexlify("d9010283"))


@pytest.mark.parametrize(
    "frozen",
    [
        pytest.param(False, id="set"),
        pytest.param(True, id="frozenset"),
    ],
)
def test_canonical_set(frozen: bool) -> None:
    value: Set[str] = {"y", "x", "aa", "a"}
    if frozen:
        value = frozenset(value)

    serialized = dumps(value, canonical=True)
    assert serialized == unhexlify("d9010284616161786179626161")


@pytest.mark.parametrize(
    "value",
    [
        pytest.param("", id="empty string"),
        pytest.param("a", id="short string"),
        pytest.param("abcde", id="long string"),
        pytest.param(b"\x01\x02\x03\x04", id="bytestring"),
        pytest.param(["a", "bb", "a", "bb"], id="array of short strings"),
        pytest.param(["a", "bb", "ccc", "dddd", "a", "bb"], id="no repeated long strings"),
        pytest.param(
            {"a": "m", "bb": "nn", "e": "m", "ff": "nn"}, id="dict with short keys and strings"
        ),
        pytest.param(
            {"a": "m", "bb": "nn", "ccc": "ooo", "dddd": "pppp", "e": "m", "ff": "nn"},
            id="dict with no repeated long strings",
        ),
    ],
)
def test_encode_stringrefs_unchanged(value: object) -> None:
    expected = dumps(value)
    if isinstance(value, list) or isinstance(value, dict):
        expected = b"\xd9\x01\x00" + expected

    assert dumps(value, string_referencing=True) == expected


def test_encode_stringrefs_array() -> None:
    value = ["aaaa", "aaaa", "bbbb", "aaaa", "bbbb"]
    equivalent = [
        "aaaa",
        CBORTag(25, 0),
        "bbbb",
        CBORTag(25, 0),
        CBORTag(25, 1),
    ]
    assert dumps(value, string_referencing=True) == b"\xd9\x01\x00" + dumps(equivalent)


def test_encode_stringrefs_dict() -> None:
    value = {"aaaa": "mmmm", "bbbb": "bbbb", "cccc": "aaaa", "mmmm": "aaaa"}
    expected = unhexlify(
        "d90100a46461616161646d6d6d6d6462626262d819026463636363d81900d81901d81900"
    )
    assert dumps(value, string_referencing=True, canonical=True) == expected


def test_encode_stringrefs_datetime() -> None:
    value = [datetime(2026, 1, 19, tzinfo=timezone.utc), "abc", "abc"]
    expected = unhexlify("D9010083C074323032362D30312D31395430303A30303A30305A63616263D81901")
    assert dumps(value, string_referencing=True) == expected


@pytest.mark.parametrize(
    "tag",
    [
        pytest.param(-1, id="too small"),
        pytest.param(2**64, id="too large"),
        pytest.param("f", id="wrong type"),
    ],
)
def test_invalid_tag(tag: object) -> None:
    with pytest.raises(TypeError):
        dumps(CBORTag(tag, "value"))  # type: ignore[arg-type]


def test_largest_tag() -> None:
    expected = unhexlify("dbffffffffffffffff6176")
    assert dumps(CBORTag(2**64 - 1, "v")) == expected


@given(compound_types_strategy)
def test_invariant_encode_decode(val: object) -> None:
    """
    Tests that an encode and decode is invariant (the value is the same after
    undergoing an encode and decode)
    """
    assert loads(dumps(val)) == val


def test_indefinite_containers() -> None:
    expected = b"\x82\x00\x01"
    assert dumps([0, 1]) == expected

    expected = b"\x9f\x00\x01\xff"
    assert dumps([0, 1], indefinite_containers=True) == expected
    assert dumps([0, 1], indefinite_containers=True, canonical=True) == expected

    expected = b"\xa0"
    assert dumps({}) == expected

    expected = b"\xbf\xff"
    assert dumps({}, indefinite_containers=True) == expected
    assert dumps({}, indefinite_containers=True, canonical=True) == expected


class TestEncoderReuse:
    """
    Tests for correct behavior when reusing CBOREncoder instances.
    """

    def test_encoder_reuse_resets_shared_containers(self) -> None:
        """
        Shared container tracking should be scoped to a single encode operation,
        not persist across multiple encodes on the same encoder instance.
        """
        fp = BytesIO()
        encoder = CBOREncoder(fp, value_sharing=True)
        shared_obj = ["hello"]

        # First encode: object is tracked in shared containers
        encoder.encode([shared_obj, shared_obj])

        # Second encode on new fp: should produce valid standalone CBOR
        # (not a sharedref pointing to stale first-encode data)
        encoder.fp = fp = BytesIO()
        encoder.encode(shared_obj)
        second_output = fp.getvalue()

        # The second output must be decodable on its own
        result = loads(second_output)
        assert result == ["hello"]

    def test_encode_to_bytes_resets_shared_containers(self) -> None:
        """
        encode_to_bytes should also reset shared container tracking between calls.
        """
        fp = BytesIO()
        encoder = CBOREncoder(fp, value_sharing=True)
        shared_obj = ["hello"]

        # First encode
        result_bytes = encoder.encode_to_bytes([shared_obj, shared_obj])
        assert result_bytes == unhexlify("d81c82d81c816568656c6c6fd81d01")

        # Second encode should produce valid standalone CBOR
        result_bytes = encoder.encode_to_bytes(shared_obj)
        result = loads(result_bytes)
        assert result == ["hello"]

    def test_encoder_hook_does_not_reset_state(self) -> None:
        """
        When a custom encoder hook calls encode(), the shared container
        tracking should be preserved (not reset mid-operation).
        """

        class Custom:
            def __init__(self, value: object):
                self.value = value

        def custom_encoder(encoder: CBOREncoder, obj: Any) -> None:
            # Hook encodes the wrapped value
            encoder.encode(obj.value)

        # Encode a Custom wrapping a list
        data = dumps(Custom(["a", "b"]), default=custom_encoder)

        # Verify the output decodes correctly
        result = loads(data)
        assert result == ["a", "b"]

        # Test nested Custom objects - hook should work recursively
        data2 = dumps(Custom(Custom(["x"])), default=custom_encoder)
        result2 = loads(data2)
        assert result2 == ["x"]



import sys

import atheris

from cbor2 import loads


def test_one_input(data: bytes) -> None:
    try:
        loads(data)
    except Exception:
        # We're searching for memory corruption, not Python exceptions
        pass


if __name__ == "__main__":
    atheris.Setup(sys.argv, test_one_input)
    atheris.Fuzz()
